import unittest
import logging
import pika
import esgfpid.rabbit.rabbit
import time


# Logging
LOGGER = logging.getLogger(__name__)
LOGGER.addHandler(logging.NullHandler())

# Test resources:
from tests.resources.IGNORE.TESTVALUES_IGNORE import TESTVALUES_REAL

class RabbitIntegrationTestCase(unittest.TestCase):

    def setUp(self):
        LOGGER.info('######## Next test ##########')

    def tearDown(self):
        LOGGER.info('#############################')

    def make_rabbit_with_access(self):
        testrabbit = esgfpid.rabbit.rabbit.RabbitMessageSender(
            exchange_name=TESTVALUES_REAL['exchange_name'],
            urls_fallback=TESTVALUES_REAL['url_messaging_service'],
            url_preferred=None,
            username=TESTVALUES_REAL['rabbit_username'],
            password=TESTVALUES_REAL['rabbit_password'],
        )
        return testrabbit

    def make_rabbit_with_several_urls(self):
        testrabbit = esgfpid.rabbit.rabbit.RabbitMessageSender(
            exchange_name=TESTVALUES_REAL['exchange_name'],
            #urls_fallback=['bar', TESTVALUES_REAL['url_messaging_service']],
            urls_fallback=['bar', 'baz'],
            url_preferred='foo',
            username=TESTVALUES_REAL['rabbit_username'],
            password=TESTVALUES_REAL['rabbit_password'],
        )
        return testrabbit

    # Tests


    if True: # WORKS
        def test_send_10_messages_to_queue_after_a_while_ok(self):
            '''
            This tests if messages are published correctly
            under normal conditions.

            The messages are sent after the thread had enough
            time to build the connection to the rabbit.
            '''

            # Test variables:
            routing_key = TESTVALUES_REAL['routing_key']
            num_messages = 10

            # Make connection:
            testrabbit = self.make_rabbit_with_access()
            testrabbit.start()

            # Wait for the connection to be established:
            time.sleep(5)

            # Run code to be tested:
            for i in xrange(num_messages):
                testrabbit.send_message_to_queue({"stuffy":"foo", "routing_key:": routing_key, "num":str(i+1)})

            # Close connection:
            LOGGER.debug('Sent. Closing now.')
            testrabbit.finish()

            # Check result:
            leftovers = testrabbit.get_leftovers()
            self.assertTrue(len(leftovers)==0, 'There is leftovers: '+str(len(leftovers)))
            self.assertTrue(testrabbit.is_finished(), 'Thread is still alive.')

    if True: # WORKS
        def test_several_urls_none_works(self):
            '''
            This tests if messages are published correctly
            under normal conditions.

            The messages are sent after the thread had enough
            time to build the connection to the rabbit.
            '''

            # Test variables:
            num_messages = 10
            routing_key = TESTVALUES_REAL['routing_key']
            message_body = '{"stuff":"foo"}'

            # Make connection:
            testrabbit = self.make_rabbit_with_several_urls()
            testrabbit.start()

            # Wait for the connection to be established:
            time.sleep(5)

            # Run code to be tested:
            for i in xrange(num_messages):
                testrabbit.send_message_to_queue({"stuffo":"foo","num":str(i+1)})

            # Close connection:
            testrabbit.finish()

            # Check result:
            leftovers = testrabbit.get_leftovers()
            self.assertTrue(len(leftovers)==num_messages, 'No message should have been sent, but there is '+str(len(leftovers))+' instead of '+str(num_messages))
            self.assertTrue(testrabbit.is_finished(), 'Thread is still alive.')

    if True:

        def test_send_10_messages_to_queue_quickly_ok(self):
            '''
            This tests if messages which cannot be sent immediately
            (because the connection is not ready) are sent later.

            The messages are published immediately after init of
            the sender, when the chain to build the connection is
            not ready yet.
            The method __check_for_already_arrived_messages() should
            find and publish them once the connection is built.
            '''

            # Test variables:
            routing_key = TESTVALUES_REAL['routing_key']
            num_messages = 10

            # Make connection:
            testrabbit = self.make_rabbit_with_access()
            testrabbit.start()

            # Run code to be tested:
            for i in xrange(num_messages):
                testrabbit.send_message_to_queue({"stuffee":"foo","num":str(i+1)})

            # Close connection after some time:
            time.sleep(5)
            testrabbit.finish()

            # Check result:
            leftovers = testrabbit.get_leftovers()
            self.assertTrue(len(leftovers)==0, 'There is leftovers: '+str(len(leftovers)))
            self.assertTrue(testrabbit.is_finished(), 'Thread is still alive.')

    if True:

        def test_send_100_messages_to_queue_and_force_finish(self):

            # Test variables:
            routing_key = TESTVALUES_REAL['routing_key']
            num_messages = 100

            # Make connection:
            testrabbit = self.make_rabbit_with_access()
            testrabbit.start()
            time.sleep(0.1)

            # Run code to be tested:
            for i in xrange(num_messages):
                testrabbit.send_message_to_queue({"stuff":"foo","num":str(i+1)})

            # Close connection after some time:
            testrabbit.force_finish()

            # Check result:
            leftovers = testrabbit.get_leftovers()
            self.assertTrue(len(leftovers)>0, 'There is no leftovers! There should be!')
            self.assertTrue(testrabbit.is_finished(), 'Thread is still alive.')



import unittest
import mock
import logging
import json
import sys
sys.path.append("..")
import tests.mocks.solrmock
import tests.mocks.rabbitmock
import esgfpid

# Logging
LOGGER = logging.getLogger(__name__)
LOGGER.addHandler(logging.NullHandler())

# Test resources:
from resources.TESTVALUES import TESTVALUES as TESTVALUES


class ConsistencyTestCase(unittest.TestCase):

    def setUp(self):
        LOGGER.info('######## Next test ##########')

    def tearDown(self):
        LOGGER.info('#############################')

    def __get_args_dict(self, testcoupler):
        return dict(
            drs_id = TESTVALUES['drs_id1'],
            data_node = TESTVALUES['data_node'],
            version_number = TESTVALUES['version_number1'],
            coupler = testcoupler
        )

    def __make_patched_testcoupler(self, solr_off=False):
        testcoupler = esgfpid.coupling.Coupler(
            handle_prefix = TESTVALUES['prefix'],
            messaging_service_urls = 'rabbit_should_not_be_used',
            messaging_service_url_preferred = None,
            messaging_service_exchange_name = TESTVALUES['messaging_exchange'],
            messaging_service_username = TESTVALUES['rabbit_username'],
            messaging_service_password = TESTVALUES['rabbit_password'],
            data_node = TESTVALUES['data_node'],
            thredds_service_path = TESTVALUES['thredds_service_path'],
            solr_url = 'solr_should_not_be_used',
            solr_https_verify=True,
            solr_switched_off=solr_off
        )
        return testcoupler

    def __patch_testcoupler_with_solr_mock(self, testcoupler, previous_files, raise_error=None):
        if raise_error is None:
            solrmock = tests.mocks.solrmock.MockSolrInteractor(previous_files=previous_files)
        else:
            solrmock = tests.mocks.solrmock.MockSolrInteractor(previous_files=previous_files, raise_error=raise_error)
        testcoupler._Coupler__solr_sender = solrmock

    def test_preparation_ok(self):

        # Prepare
        testcoupler = self.__make_patched_testcoupler()
        prev = ['abc', 'def']
        self.__patch_testcoupler_with_solr_mock(testcoupler, prev)
        args = self.__get_args_dict(testcoupler)

        # Run code to be tested:
        checker = esgfpid.assistant.consistency.Checker(**args)

        # Check result:
        self.assertIsInstance(checker, esgfpid.assistant.consistency.Checker,
            'Preparation failed.')
        self.assertTrue(checker.can_run_check(),
            'With previous files, check should be enabled, but isn\'t.')

    def test_check_ok(self):

        # Prepare
        testcoupler = self.__make_patched_testcoupler()
        prev = ['abc', 'def']
        self.__patch_testcoupler_with_solr_mock(testcoupler, prev)
        args = self.__get_args_dict(testcoupler)
        checker = esgfpid.assistant.consistency.Checker(**args)

        # Run code to be tested:
        success = checker.data_consistency_check(prev)

        # Check result:
        self.assertTrue(success, 'The check should pass.')

    def test_check_too_many_files(self):

        # Prepare
        testcoupler = self.__make_patched_testcoupler()
        prev = ['abc', 'def']
        too_many_files = prev+['ghi', 'jkl']
        self.__patch_testcoupler_with_solr_mock(testcoupler, prev)
        args = self.__get_args_dict(testcoupler)
        checker = esgfpid.assistant.consistency.Checker(**args)

        # Run code to be tested:
        success = checker.data_consistency_check(too_many_files)

        # Check result:
        self.assertFalse(success, 'The check should fail - too many files passed.')

    def test_check_too_few_files(self):

        # Prepare
        testcoupler = self.__make_patched_testcoupler()
        prev = ['abc', 'def']
        too_few_files = ['abc']
        self.__patch_testcoupler_with_solr_mock(testcoupler, prev)
        args = self.__get_args_dict(testcoupler)
        checker = esgfpid.assistant.consistency.Checker(**args)

        # Run code to be tested:
        success = checker.data_consistency_check(too_few_files)

        # Check result:
        self.assertFalse(success, 'The check should fail - too few files passed.')

    def test_check_different_files(self):

        # Prepare
        testcoupler = self.__make_patched_testcoupler()
        prev = ['abc', 'def']
        different_files = ['xyz', 'zyx']
        self.__patch_testcoupler_with_solr_mock(testcoupler, prev)
        args = self.__get_args_dict(testcoupler)
        checker = esgfpid.assistant.consistency.Checker(**args)

        # Run code to be tested:
        success = checker.data_consistency_check(different_files)

        # Check result:
        self.assertFalse(success, 'The check should fail - different files passed.')

    def test_check_no_files(self):

        # Prepare
        testcoupler = self.__make_patched_testcoupler()
        prev = ['abc', 'def']
        no_files = []
        self.__patch_testcoupler_with_solr_mock(testcoupler, prev)
        args = self.__get_args_dict(testcoupler)
        checker = esgfpid.assistant.consistency.Checker(**args)

        # Run code to be tested:
        with self.assertRaises(ValueError):
            success = checker.data_consistency_check(no_files)

    def test_check_try_if_disabled(self):

        # Prepare
        testcoupler = self.__make_patched_testcoupler()
        self.__patch_testcoupler_with_solr_mock(testcoupler, 'any_previous_files')
        args = self.__get_args_dict(testcoupler)
        checker = esgfpid.assistant.consistency.Checker(**args)
        checker._Checker__will_run_check = False # Switch the check off!

        # Run code to be tested:
        with self.assertRaises(ValueError) as raised:
            success = checker.data_consistency_check(['any', 'thing'])
        self.assertIn('No reason specified', raised.exception.message, 'Expected message did not arrive: "%s"' % raised.exception.message)


    def test_preparation_files_is_empty_list(self):

        # Prepare
        testcoupler = self.__make_patched_testcoupler()
        prev = []
        self.__patch_testcoupler_with_solr_mock(testcoupler, prev)
        args = self.__get_args_dict(testcoupler)

        # Run code to be tested:
        checker = esgfpid.assistant.consistency.Checker(**args)

        # Check result:
        self.assertIsInstance(checker, esgfpid.assistant.consistency.Checker,
            'Preparation failed.')
        self.assertFalse(checker.can_run_check(),
            'Without previous files, check should be disabled, but isn\'t.')

    def test_preparation_files_is_none(self):

        # Prepare
        testcoupler = self.__make_patched_testcoupler()
        self.__patch_testcoupler_with_solr_mock(testcoupler, 'NONE') # makes the mock return None. In reality, this should never happen!
        args = self.__get_args_dict(testcoupler)

        # Run code to be tested:
        checker = esgfpid.assistant.consistency.Checker(**args)

        # Check result:
        self.assertIsInstance(checker, esgfpid.assistant.consistency.Checker,
            'Preparation failed.')
        self.assertFalse(checker.can_run_check(),
            'Without previous files, check should be disabled, but isn\'t.')

    def test_preparation_error(self):

        # Prepare
        testcoupler = self.__make_patched_testcoupler()
        error_to_raise = esgfpid.exceptions.SolrError
        self.__patch_testcoupler_with_solr_mock(testcoupler, None, error_to_raise) # makes the mock raise an error
        args = self.__get_args_dict(testcoupler)

        # Run code to be tested:
        checker = esgfpid.assistant.consistency.Checker(**args)

        # Check result:
        self.assertIsInstance(checker, esgfpid.assistant.consistency.Checker,
            'Preparation failed.')
        self.assertFalse(checker.can_run_check(),
            'Without previous files, check should be disabled, but isn\'t.')

    def test_preparation_switched_off(self):

        # Prepare
        testcoupler = self.__make_patched_testcoupler(solr_off=True)
        args = self.__get_args_dict(testcoupler)

        # Run code to be tested:
        checker = esgfpid.assistant.consistency.Checker(**args)

        # Check result:
        self.assertIsInstance(checker, esgfpid.assistant.consistency.Checker,
            'Preparation failed.')
        self.assertFalse(checker.can_run_check(),
            'Without previous files, check should be disabled, but isn\'t.')




expected_message_ok = '''Checking config for PID module (rabbit messaging queue) ...
 .. checking authentication and connection ...
 .. checking authentication and connection (this.is.my.only.host)... ok.
 .. checking authentication and connection ... ok.
 .. checking channel ...
 .. checking channel ... ok.
Config for PID module (rabbit messaging queue).. ok.
Successful connection to PID messaging queue at "this.is.my.only.host".'''


expected_message_first_connection_failed_then_ok = '''Checking config for PID module (rabbit messaging queue) ...
 .. checking authentication and connection ...
 .. checking connection (this.is.my.favourite.host)... FAILED.
 .. checking connection (mystery-tour.uk)... FAILED.
 .. checking authentication and connection (tomato.salad-with-spam.fr)... ok.
 .. checking authentication and connection ... ok.
 .. checking channel ...
 .. checking channel ... ok.
Config for PID module (rabbit messaging queue).. ok.
Successful connection to PID messaging queue at "tomato.salad-with-spam.fr".'''


expected_message_connection_failed_unknown = '''Checking config for PID module (rabbit messaging queue) ...
 .. checking authentication and connection ...
 .. checking connection (this.is.my.only.host)... FAILED.
Config for PID module (rabbit messaging queue) .. FAILED!
********************************************************************
*** PROBLEM IN SETTING UP                                        ***
*** RABBIT MESSAGING QUEUE (PID MODULE)                          ***
*** CONNECTION TO THE PID MESSAGING QUEUE FAILED DEFINITIVELY:   ***
***  - host "this.is.my.only.host": Unknown connection failure.  ***
*** PLEASE NOTIFY handle@dkrz.de AND INCLUDE THIS ERROR MESSAGE. ***
********************************************************************'''


expected_message_connection_failed_only = '''Checking config for PID module (rabbit messaging queue) ...
 .. checking authentication and connection ...
 .. checking connection (this.is.my.only.host)... FAILED.
Config for PID module (rabbit messaging queue) .. FAILED!
********************************************************************
*** PROBLEM IN SETTING UP                                        ***
*** RABBIT MESSAGING QUEUE (PID MODULE)                          ***
*** CONNECTION TO THE PID MESSAGING QUEUE FAILED DEFINITIVELY:   ***
***  - host "this.is.my.only.host": Connection failure.          ***
*** PLEASE NOTIFY handle@dkrz.de AND INCLUDE THIS ERROR MESSAGE. ***
********************************************************************'''

expected_message_connection_failed_several = '''Checking config for PID module (rabbit messaging queue) ...
 .. checking authentication and connection ...
 .. checking connection (this.is.my.favourite.host)... FAILED.
 .. checking connection (mystery-tour.uk)... FAILED.
 .. checking connection (tomato.salad-with-spam.fr)... FAILED.
Config for PID module (rabbit messaging queue) .. FAILED!
********************************************************************
*** PROBLEM IN SETTING UP                                        ***
*** RABBIT MESSAGING QUEUE (PID MODULE)                          ***
*** CONNECTION TO THE PID MESSAGING QUEUE FAILED DEFINITIVELY:   ***
***  - host "this.is.my.favourite.host": Connection failure.     ***
***  - host "mystery-tour.uk": Connection failure.               ***
***  - host "tomato.salad-with-spam.fr": Connection failure.     ***
*** PLEASE NOTIFY handle@dkrz.de AND INCLUDE THIS ERROR MESSAGE. ***
********************************************************************'''

expected_message_authentication_failed_only = '''Checking config for PID module (rabbit messaging queue) ...
 .. checking authentication and connection ...
 .. checking authentication (this.is.my.only.host)... FAILED.
Config for PID module (rabbit messaging queue) .. FAILED!
*************************************************************************************************
*** PROBLEM IN SETTING UP                                                                     ***
*** RABBIT MESSAGING QUEUE (PID MODULE)                                                       ***
*** CONNECTION TO THE PID MESSAGING QUEUE FAILED DEFINITIVELY:                                ***
***  - host "this.is.my.only.host": Authentication failure (user johndoe, password abc123yx). ***
*** PLEASE NOTIFY handle@dkrz.de AND INCLUDE THIS ERROR MESSAGE.                              ***
*************************************************************************************************'''

expected_message_authentication_failed_several = '''Checking config for PID module (rabbit messaging queue) ...
 .. checking authentication and connection ...
 .. checking authentication (this.is.my.favourite.host)... FAILED.
 .. checking authentication (mystery-tour.uk)... FAILED.
 .. checking authentication (tomato.salad-with-spam.fr)... FAILED.
Config for PID module (rabbit messaging queue) .. FAILED!
******************************************************************************************************
*** PROBLEM IN SETTING UP                                                                          ***
*** RABBIT MESSAGING QUEUE (PID MODULE)                                                            ***
*** CONNECTION TO THE PID MESSAGING QUEUE FAILED DEFINITIVELY:                                     ***
***  - host "this.is.my.favourite.host": Authentication failure (user johndoe, password abc123yx). ***
***  - host "mystery-tour.uk": Authentication failure (user johndoe, password abc123yx).           ***
***  - host "tomato.salad-with-spam.fr": Authentication failure (user johndoe, password abc123yx). ***
*** PLEASE NOTIFY handle@dkrz.de AND INCLUDE THIS ERROR MESSAGE.                                   ***
******************************************************************************************************'''


expected_message_channel_closed_only = '''Checking config for PID module (rabbit messaging queue) ...
 .. checking authentication and connection ...
 .. checking authentication and connection (this.is.my.only.host)... ok.
 .. checking authentication and connection ... ok.
 .. checking channel ...
 .. checking channel ... FAILED.
Config for PID module (rabbit messaging queue) .. FAILED!
********************************************************************
*** PROBLEM IN SETTING UP                                        ***
*** RABBIT MESSAGING QUEUE (PID MODULE)                          ***
*** CONNECTION TO THE PID MESSAGING QUEUE FAILED DEFINITIVELY:   ***
***  - host "this.is.my.only.host": Channel failure.             ***
*** PLEASE NOTIFY handle@dkrz.de AND INCLUDE THIS ERROR MESSAGE. ***
********************************************************************'''


expected_message_first_channel_closed_then_ok = '''Checking config for PID module (rabbit messaging queue) ...
 .. checking authentication and connection ...
 .. checking authentication and connection (this.is.my.favourite.host)... ok.
 .. checking authentication and connection ... ok.
 .. checking channel ...
 .. checking channel ... FAILED.
 .. checking authentication and connection ...
 .. checking authentication and connection (mystery-tour.uk)... ok.
 .. checking authentication and connection ... ok.
 .. checking channel ...
 .. checking channel ... FAILED.
 .. checking authentication and connection ...
 .. checking authentication and connection (tomato.salad-with-spam.fr)... ok.
 .. checking authentication and connection ... ok.
 .. checking channel ...
 .. checking channel ... ok.
Config for PID module (rabbit messaging queue).. ok.
Successful connection to PID messaging queue at "tomato.salad-with-spam.fr".'''


import unittest
import mock
import logging
import json
import sys
import esgfpid.assistant.unpublish
import tests.mocks.rabbitmock
import tests.mocks.solrmock
import tests.utils
from esgfpid.defaults import ROUTING_KEY_BASIS as ROUTING_KEY_BASIS

LOGGER = logging.getLogger(__name__)
LOGGER.addHandler(logging.NullHandler())

class UnpublicationTestCase(unittest.TestCase):

    def tearDown(self):
        LOGGER.info('#############################')

    def setUp(self):
        LOGGER.info('######## Next test (%s) ##########', __name__)
        self.prefix = '123456'
        self.data_node = 'my.data.node'
        self.drs_id = 'mytest'
        self.coupler = self.make_coupler_with_defaults_and_mocks()

    @mock.patch('esgfpid.rabbit.rabbit.RabbitMessageSender.__init__', mock.Mock(return_value=None))
    @mock.patch('esgfpid.solr.solr.SolrInteractor.__init__', mock.Mock(return_value=None))
    def make_coupler_with_defaults_and_mocks(self):

        # Make coupler
        args = self.get_args_for_coupler()
        testcoupler = esgfpid.coupling.Coupler(**args)

        # Replace objects that interact with servers with mocks
        self.__patch_coupler_with_solr_mock(testcoupler)
        self.__patch_coupler_with_rabbit_mock(testcoupler)
        return testcoupler

    def __patch_coupler_with_solr_mock(self, coupler, solrmock=None):
        if solrmock is None:
            solrmock = tests.mocks.solrmock.MockSolrInteractor()
        coupler._Coupler__solr_sender = solrmock

    def __patch_coupler_with_rabbit_mock(self, coupler, rabbitmock=None):
        if rabbitmock is None:
            rabbitmock = tests.mocks.rabbitmock.SimpleMockRabbitSender()
        coupler._Coupler__rabbit_message_sender = rabbitmock

    def __get_received_message_from_rabbit_mock(self, coupler, index):
        return coupler._Coupler__rabbit_message_sender.received_messages[index]

    def get_args_for_coupler(self):
        args = dict(
            handle_prefix=self.prefix,
            messaging_service_urls_open='www.rabbit.foo',
            messaging_service_url_trusted='www.trusted-rabbit.foo',
            messaging_service_exchange_name='exch',
            messaging_service_username_open='rogerRabbit',
            messaging_service_username_trusted='rogerRabbit',
            messaging_service_password='mySecretCarrotDream',
            solr_url='foo.solr.bar',
            solr_https_verify=False,
            solr_switched_off=True
        )
        return args

    def make_assistant_one_version(self):
        assistant = esgfpid.assistant.unpublish.AssistantOneVersion(
            drs_id=self.drs_id,
            prefix=self.prefix,
            coupler=self.coupler,
            data_node=self.data_node,
            message_timestamp='some_moment'
        )
        return assistant

    def make_assistant_all_versions(self, coupler=None):
        assistant = esgfpid.assistant.unpublish.AssistantAllVersions(
            drs_id=self.drs_id,
            prefix=self.prefix,
            coupler=coupler or self.coupler,
            data_node=self.data_node,
            message_timestamp='some_moment',
            consumer_solr_url='fake_solr_foo'
        )
        return assistant

    # Tests

    def test_unpublish_one_version_by_version_number(self):

        # Test variables
        version_number='999888'

        # Preparations
        assistant = self.make_assistant_one_version()
  
        # Run code to be tested:
        assistant.unpublish_one_dataset_version(
            version_number=version_number
        )

        # Check result:
        expected_rabbit_task = {
            "handle": "hdl:"+self.prefix+"/8bce9b3c-88eb-3f6d-a409-4f69ef7d043f",
            "aggregation_level": "dataset",
            "operation": "unpublish_one_version",
            "message_timestamp":"anydate",
            "data_node": self.data_node,
            "ROUTING_KEY": ROUTING_KEY_BASIS+'unpublication.one',
            "drs_id":self.drs_id,
            "version_number":int(version_number)
        }
        received_rabbit_task = self.__get_received_message_from_rabbit_mock(self.coupler, 0)

        messagesOk = tests.utils.compare_json_return_errormessage(expected_rabbit_task, received_rabbit_task)
        self.assertIsNone(messagesOk, messagesOk)

    def test_unpublish_one_version_by_version_number_and_handle(self):

        # Test variables
        version_number='999888'
        handle = 'hdl:'+self.prefix + '/8bce9b3c-88eb-3f6d-a409-4f69ef7d043f' # is redundant, but will be checked.
  
        # Preparations
        assistant = self.make_assistant_one_version()

        # Run code to be tested:
        assistant.unpublish_one_dataset_version(
            version_number=version_number,
            dataset_handle=handle
        )

        # Check result:
        expected_rabbit_task = {
            "handle": "hdl:"+self.prefix+"/8bce9b3c-88eb-3f6d-a409-4f69ef7d043f",
            "aggregation_level": "dataset",
            "operation": "unpublish_one_version",
            "message_timestamp":"anydate",
            "data_node": self.data_node,
            "ROUTING_KEY": ROUTING_KEY_BASIS+'unpublication.one',
            "drs_id":self.drs_id,
            "version_number":int(version_number)
        }
        received_rabbit_task = self.__get_received_message_from_rabbit_mock(self.coupler, 0)
        messagesOk = tests.utils.compare_json_return_errormessage(expected_rabbit_task, received_rabbit_task)
        self.assertIsNone(messagesOk, messagesOk)

    def test_unpublish_one_version_wrong_handle(self):

        # Test variables
        version_number = '999888'
        wrong_handle = self.prefix+'/miauz'
  
        # Preparations
        assistant = self.make_assistant_one_version()

        # Run code to be tested and check exception:
        with self.assertRaises(ValueError):
            assistant.unpublish_one_dataset_version(version_number=version_number, dataset_handle=wrong_handle)

    def test_unpublish_one_version_version_is_none(self):
        
        # Preparations
        assistant = self.make_assistant_one_version()

        # Run code to be tested and check exception:
        with self.assertRaises(esgfpid.exceptions.ArgumentError):
            assistant.unpublish_one_dataset_version(version_number=None)

    def test_unpublish_one_version_no_version_given(self):
        
        # Preparations
        assistant = self.make_assistant_one_version()

        # Run code to be tested and check exception:
        with self.assertRaises(esgfpid.exceptions.ArgumentError):
            assistant.unpublish_one_dataset_version()

    def test_unpublish_all_versions_nosolr_consumer_must_find_versions_ok(self):        

        # Preparations:
        assistant = self.make_assistant_all_versions()

        # Run code to be tested:
        assistant.unpublish_all_dataset_versions()

        # Check result:
        expected_rabbit_task = {
            "operation": "unpublish_all_versions",
            "aggregation_level": "dataset",
            "message_timestamp": "anydate",
            "drs_id": self.drs_id,
            "data_node": self.data_node,
            "ROUTING_KEY": ROUTING_KEY_BASIS+'unpublication.all',
            "consumer_solr_url":"fake_solr_foo"
        }
        received_rabbit_task = self.__get_received_message_from_rabbit_mock(self.coupler, 0)
        messagesOk = tests.utils.compare_json_return_errormessage(expected_rabbit_task, received_rabbit_task)
        self.assertIsNone(messagesOk, messagesOk)

    def test_unpublish_all_versions_solr_off_consumer_must_find_versions_ok(self):        

        # Preparations:
        # Make coupler
        args = self.get_args_for_coupler()
        args['solr_url']=None
        args['solr_switched_off']=True
        testcoupler = esgfpid.coupling.Coupler(**args)
        # Replace objects that interact with servers with mocks
        self.__patch_coupler_with_rabbit_mock(testcoupler)
        # Make assistant
        assistant = self.make_assistant_all_versions(testcoupler)

        # Run code to be tested:
        assistant.unpublish_all_dataset_versions()

        # Check result:
        expected_rabbit_task = {
            "operation": "unpublish_all_versions",
            "aggregation_level": "dataset",
            "message_timestamp": "anydate",
            "drs_id": self.drs_id,
            "data_node": self.data_node,
            "ROUTING_KEY": ROUTING_KEY_BASIS+'unpublication.all',
            "consumer_solr_url":"fake_solr_foo"
        }
        received_rabbit_task = self.__get_received_message_from_rabbit_mock(testcoupler, 0)
        messagesOk = tests.utils.compare_json_return_errormessage(expected_rabbit_task, received_rabbit_task)
        self.assertIsNone(messagesOk, messagesOk)

    def test_unpublish_all_versions_nosolr_consumer_must_find_versions_solr_none(self):        

        # Preparations:
        assistant = esgfpid.assistant.unpublish.AssistantAllVersions(
            drs_id=self.drs_id,
            prefix=self.prefix,
            coupler=self.coupler,
            data_node=self.data_node,
            message_timestamp='some_moment',
            consumer_solr_url=None
        )

        # Run code to be tested:
        assistant.unpublish_all_dataset_versions()

        # Check result:
        expected_rabbit_task = {
            "operation": "unpublish_all_versions",
            "aggregation_level": "dataset",
            "message_timestamp": "anydate",
            "drs_id": self.drs_id,
            "data_node": self.data_node,
            "ROUTING_KEY": ROUTING_KEY_BASIS+'unpublication.all'
        }
        received_rabbit_task = self.__get_received_message_from_rabbit_mock(self.coupler, 0)
        messagesOk = tests.utils.compare_json_return_errormessage(expected_rabbit_task, received_rabbit_task)
        self.assertIsNone(messagesOk, messagesOk)

    def test_unpublish_all_versions_nosolr_consumer_must_find_versions_no_solr(self):        

        # Check exception:

        assistant = esgfpid.assistant.unpublish.AssistantAllVersions(
            drs_id=self.drs_id,
            prefix=self.prefix,
            coupler=self.coupler,
            data_node=self.data_node,
            message_timestamp='some_moment',
        )

    def test_unpublish_all_versions_by_handles_ok(self):

        # Test variables
        list_of_handles = [self.prefix+'/bla', self.prefix+'/blub']

        # Set solr mock to return handles:
        self.coupler._Coupler__solr_sender.datasethandles_or_versionnumbers_of_allversions['dataset_handles'] = list_of_handles
  
        # Preparations:
        assistant = self.make_assistant_all_versions()

        # Run code to be tested:
        assistant.unpublish_all_dataset_versions()

        # Check result:
        expected_rabbit_task_0 = {
            "operation": "unpublish_one_version",
            "aggregation_level": "dataset",
            "message_timestamp": "anydate",
            "data_node": self.data_node,
            "handle":list_of_handles[0],
            "ROUTING_KEY": ROUTING_KEY_BASIS+'unpublication.one',
            "drs_id": self.drs_id
        }
        expected_rabbit_task_1 = {
            "operation": "unpublish_one_version",
            "aggregation_level": "dataset",
            "message_timestamp": "anydate",
            "data_node": self.data_node,
            "handle":list_of_handles[1],
            "ROUTING_KEY": ROUTING_KEY_BASIS+'unpublication.one',
            "drs_id": self.drs_id
        }
        received_rabbit_task_0 = self.__get_received_message_from_rabbit_mock(self.coupler, 0)
        received_rabbit_task_1 = self.__get_received_message_from_rabbit_mock(self.coupler, 1)

        messagesOk = tests.utils.compare_json_return_errormessage(expected_rabbit_task_0, received_rabbit_task_0)
        self.assertIsNone(messagesOk, messagesOk)

        messagesOk = tests.utils.compare_json_return_errormessage(expected_rabbit_task_1, received_rabbit_task_1)
        self.assertIsNone(messagesOk, messagesOk)

    def test_unpublish_all_versions_by_version_numbers_ok(self):
        
        # Test variables
        list_of_version_numbers = [12345, 54321]

        # Set solr mock to return handles:
        self.coupler._Coupler__solr_sender.datasethandles_or_versionnumbers_of_allversions['version_numbers'] = list_of_version_numbers
  
        # Preparations:
        assistant = self.make_assistant_all_versions()

        # Run code to be tested:
        assistant.unpublish_all_dataset_versions()

        # Check result:
        expected_rabbit_task_0 = {
            "operation": "unpublish_one_version",
            "aggregation_level": "dataset",
            "message_timestamp": "anydate",
            "data_node": self.data_node,
            "handle": "hdl:"+self.prefix+"/68f3d5ab-bddd-3701-abf4-4b24ccd54525",
            "ROUTING_KEY": ROUTING_KEY_BASIS+'unpublication.one',
            "drs_id": self.drs_id,
            "version_number": 12345
        }
        expected_rabbit_task_1 = {
            "operation": "unpublish_one_version",
            "aggregation_level": "dataset",
            "message_timestamp": "anydate",
            "data_node": self.data_node,
            "handle": "hdl:"+self.prefix+"/fca5741d-89c1-3149-a34d-c7fb37be75a8",
            "ROUTING_KEY": ROUTING_KEY_BASIS+'unpublication.one',
            "drs_id": self.drs_id,
            "version_number": 54321
        }
        received_rabbit_task_0 = self.__get_received_message_from_rabbit_mock(self.coupler, 0)
        received_rabbit_task_1 = self.__get_received_message_from_rabbit_mock(self.coupler, 1)

        messagesOk = tests.utils.compare_json_return_errormessage(expected_rabbit_task_0, received_rabbit_task_0)
        self.assertIsNone(messagesOk, messagesOk)

        messagesOk = tests.utils.compare_json_return_errormessage(expected_rabbit_task_1, received_rabbit_task_1)
        self.assertIsNone(messagesOk, messagesOk)



import logging
import datetime
import json
import pika
import time
import esgfpid.utils
import esgfpid.defaults
from . import rabbitutils

# Normal logger:
LOGGER = logging.getLogger(__name__)
LOGGER.addHandler(logging.NullHandler())

class SynchronousServerConnector(object):

    def __init__(self, **args):
        ''' In init, no connection is opened yet.'''

        # Check args
        mandatory_args = [
            'exchange_name',
            'url_preferred',
            'urls_fallback',
            'credentials'
        ]
        esgfpid.utils.check_presence_of_mandatory_args(args, mandatory_args)

        self.__store_args(args)
        self.__retrieve_defaults()

        # Other settings:
        self.__channel = None
        self.__connection = None
        self.__communication_established = False
        self.__connection_last_process_event_call = 0
        self.__logdate = datetime.datetime.now().strftime('%Y-%m-%d_%H:%M:%S')
        self.__current_rabbit_url = None
        self.__error_messages_during_init = []

    def __store_args(self, args):
        self.__exchange_name = args['exchange_name']
        self.__credentials = args['credentials']
        self.__url_preferred = args['url_preferred']
        self.__urls_fallback = args['urls_fallback']

    def __retrieve_defaults(self):
        # For sending messages:
        self.__max_tries = esgfpid.defaults.RABBIT_SYN_MAX_TRIES
        self.__timeout_milliseconds = esgfpid.defaults.RABBIT_SYN_TIMEOUT_MILLISEC


    ########################################
    ### This actually connects to rabbit ###
    ########################################

    def open_rabbit_connection(self):
        success = self.__try_connecting_to_preferred()
        if not success:
            success = self.__try_connecting_to_fallbacks()
            if not success:
                collected_errors = ' - '.join(self.__error_messages_during_init)
                LOGGER.warning('No connection possible. Errors: %s' % collected_errors)

    def __try_connecting_to_preferred(self):
        url = self.__url_preferred
        success = self.__try_connecting_to_url(url)
        return success

    def __try_connecting_to_fallbacks(self):
        urls = self.__urls_fallback
        for url in urls:
            success = self.__try_connecting_to_url(url)
            if success:
                return success
        return False

    def __try_connecting_to_url(self, url):

        connection = self.__setup_rabbit_connection(url)

        if connection is not None and connection.is_open:
            self.__connection = connection
            self.__current_rabbit_url = url
            LOGGER.debug('Succeeded opening connection to '+str(url))

            channel = self.__setup_rabbit_channel()

            if channel is not None and channel.is_open:
                LOGGER.debug('Succeeded opening channel.')
                self.__channel = channel
                self.__channel.confirm_delivery()
                self.__communication_established = True
                return True
            else:
                LOGGER.debug('Could not open channel.')

        return False

    def __setup_rabbit_connection(self, url):
        LOGGER.debug('Setting up the connection with the RabbitMQ.')
        try:
            params = esgfpid.rabbit.connparams.get_connection_parameters(self.__credentials, url)
            conn = self.__make_connection(params)
            return conn
        except pika.exceptions.ProbableAuthenticationError:
            msg = ('Problem setting up the rabbit with username "%s" and password "%s" at url %s.' %
                    (self.__credentials.username,
                     self.__credentials.password,
                     url))
            LOGGER.error(msg)
            self.__error_messages_during_init.append(msg)
            return None
        except pika.exceptions.AMQPConnectionError:
            msg = ('Problem setting up the rabbit connection to %s.' % url)
            self.__error_messages_during_init.append(msg)
            return None

    def __make_connection(self, params): # this is easy to mock during unit tests
        return pika.BlockingConnection(params)

    def __setup_rabbit_channel(self):
        try:
            return self.__make_channel()
        except pika.exceptions.ChannelClosed:
            msg = ('Problem setting up the rabbit channel.')
            self.__error_messages_during_init.append(msg)
            LOGGER.warn(msg)
            return None

    def __make_channel(self): # this is easy to mock during unit tests
        return self.__connection.channel()

    ########################
    ### Close connection ###
    ########################

    def close_rabbit_connection(self):
        if (self.__connection is not None) and (self.__connection.is_open):
            self.__connection.close()
            # From pika doc: If there are any open channels, it will attempt to
            # close them prior to fully disconnecting

    ####################
    ### Send message ###
    ####################

    def send_message_to_queue(self, message):
        self.__open_connection_if_not_open()
        routing_key, msg_string = rabbitutils.get_routing_key_and_string_message_from_message_if_possible(message)
        self.__try_sending_message_several_times(routing_key, msg_string)

    def __open_connection_if_not_open(self):
        if not self.__communication_established:
            self.open_rabbit_connection()

    def __try_sending_message_several_times(self, routing_key, message):
        delivered_after_x_times = self.__retry_x_times(routing_key, message)
        if not delivered_after_x_times:
            LOGGER.error('Message could not be sent.')

    def __retry_x_times(self, routing_key, messagebody):
        message_delivered = False
        max_tries_reached = False
        counter = 1
        while (not message_delivered) and (not max_tries_reached):

            # Try to deliver:
            #LOGGER.debug('Delivery no. '+str(counter)')
            delivered_now = self.__send_message_to_queue_once(routing_key, messagebody)

            # Check if it was delivered:
            if delivered_now:
                LOGGER.debug('Successful delivery of message with routing key "'+str(routing_key)+'"!')
                LOGGER.debug("Delivered after "+str(counter)+" times!")
                message_delivered = True
            else:
                LOGGER.debug("Not delivered, waiting "+str(self.__timeout_milliseconds)+" milliseconds before delivering it again!")
                time.sleep(self.__timeout_milliseconds/1000)

            # Increment counter, check if we have reached the max:
            if counter == self.__max_tries:
                max_tries_reached = True
                LOGGER.debug("Tried "+str(self.__max_tries)+" times, giving up!")
            counter += 1

        # Return with or without success:
        return message_delivered

    def __send_message_to_queue_once(self, routing_key, messagebody):

        props = esgfpid.rabbit.connparams.get_properties_for_message_publications()

        delivered = False
        try:
            delivered = self.__do_send_message(routing_key, messagebody, props)
            self.__avoid_connection_shutdown()
            
        except pika.exceptions.UnroutableError:
            LOGGER.error('Message could not be routed to any queue, maybe none was declared yet.')
            # TODO
        return delivered

    def __do_send_message(self, routing_key, messagebody, props): # This is easy to mock during unit tests
        return self.__channel.basic_publish(
            exchange=self.__exchange_name,
            routing_key=routing_key,
            body=messagebody,
            mandatory=esgfpid.defaults.RABBIT_MANDATORY_DELIVERY,
            properties=props
        )

    def __avoid_connection_shutdown(self):
        # Source: https://github.com/pika/pika/issues/397
        if (time.time() - self.__connection_last_process_event_call) > 6:
            self.__connection.process_data_events()
            self.__connection_last_process_event_call = time.time()



import logging
import esgfpid.utils
import esgfpid.assistant.messages
from esgfpid.utils import loginfo, logdebug, logtrace, logerror, logwarn

LOGGER = logging.getLogger(__name__)
LOGGER.addHandler(logging.NullHandler())

class ShoppingCartAssistant(object):

    def __init__(self, **args):
        mandatory_args = ['prefix', 'coupler']
        esgfpid.utils.check_presence_of_mandatory_args(args, mandatory_args)
        esgfpid.utils.check_noneness_of_mandatory_args(args, mandatory_args)
        self.__prefix = args['prefix']
        self.__coupler = args['coupler']

    def make_shopping_cart_pid(self, dict_of_drs_ids_and_pids):
        logdebug(LOGGER, 'Making a PID for a shopping cart full of datasets...')

        # Check arg
        if not type(dict_of_drs_ids_and_pids) == type(dict()):
            if type(dict_of_drs_ids_and_pids) == type([]):
                raise esgfpid.exceptions.ArgumentError('Please provide a dictionary of dataset ids and handles, not a list')
            else:
                raise esgfpid.exceptions.ArgumentError('Please provide a dictionary of dataset ids and handles')

        # Make a pid (hash on the content):
        list_of_drs_ids = dict_of_drs_ids_and_pids.keys()
        cart_handle = self.__get_handle_for_cart(list_of_drs_ids, self.__prefix)

        # Make and send message
        message = self.__make_message(cart_handle, dict_of_drs_ids_and_pids)
        self.__send_message_to_queue(message)

        # Return pid
        logdebug(LOGGER, 'Making a PID for a shopping cart full of datasets... done.')
        loginfo(LOGGER, 'Requesting to create PID for data cart (%s).', cart_handle)
        return cart_handle

    def __get_handle_for_cart(self, list_of_drs_ids, prefix):
        hash_basis = esgfpid.utils.make_sorted_lowercase_list_without_hdl(list_of_drs_ids)
        return esgfpid.utils.make_handle_from_list_of_strings(hash_basis, prefix)
        # This sorts the list, removes all "hdl:", and makes a hash

    def __make_message(self, cart_handle, dict_of_drs_ids_and_pids):
        message_timestamp = esgfpid.utils.get_now_utc_as_formatted_string()
        message = esgfpid.assistant.messages.make_shopping_cart_message(
            cart_handle = cart_handle,
            timestamp = message_timestamp,
            data_cart_content = dict_of_drs_ids_and_pids
        )
        return message

    def __send_message_to_queue(self, message):
        self.__coupler.send_message_to_queue(message)


import logging
import esgfpid.rabbit.rabbit
import esgfpid.solr.solr
import esgfpid.utils

LOGGER = logging.getLogger(__name__)
LOGGER.addHandler(logging.NullHandler())

class Coupler(object):

    def __init__(self, **args):
        self.__create_all_attributes()
        self.__create_message_sender(args)
        self.__create_solr_sender(args)

    def __create_all_attributes(self):
        self.__rabbit_message_sender = None
        self.__solr_sender = None

    def __create_message_sender(self, args):

        credentials = args['messaging_service_credentials']
        for cred in credentials:
            if 'password' not in cred:
                cred['password'] = 'jzlnL78ZpExV#_QHz'

        if 'test_publication' in args:
            test_publication = args['test_publication']
        else:
            test_publication = False

        self.__rabbit_message_sender = esgfpid.rabbit.rabbit.RabbitMessageSender(
            exchange_name=args['messaging_service_exchange_name'],       # mandatory
            credentials=args['messaging_service_credentials'],
            test_publication=test_publication
        )

    def __create_solr_sender(self, args):

        # Some use cases need no solr, so solr url is optional:
        if args['solr_url'] is None:
            args['solr_switched_off']=True

        self.__solr_sender = esgfpid.solr.solr.SolrInteractor(
            solr_url=args['solr_url'],
            prefix=args['handle_prefix'],
            https_verify=args['solr_https_verify'],
            switched_off=args['solr_switched_off']
        )

    ### Communications with rabbit

    def send_message_to_queue(self, message):
        success = self.__rabbit_message_sender.send_message_to_queue(message)
        return success

    ### For synchronous

    def start_rabbit_business(self):
        '''
        Open a connection. If not called, the first sent message automatically
        opens a connection.
        This is called for example by the publish assistant.
        '''
        self.__rabbit_message_sender.open_rabbit_connection()

    def done_with_rabbit_business(self):
        '''
        Optional.
        This is called for example by the publish assistant.
        '''
        self.__rabbit_message_sender.close_rabbit_connection()

    ### For asynchronous

    def start_rabbit_connection(self):
        self.__rabbit_message_sender.start()

    def finish_rabbit_connection(self):
        self.__rabbit_message_sender.finish()

    def force_finish_rabbit_connection(self):
        self.__rabbit_message_sender.force_finish()

    ### Communications with solr

    def retrieve_datasethandles_or_versionnumbers_of_allversions(self, **args):
        
        mandatory_args = ['drs_id']
        esgfpid.utils.check_presence_of_mandatory_args(args, mandatory_args)
        esgfpid.utils.check_noneness_of_mandatory_args(args, mandatory_args)

        result_dict = self.__solr_sender.retrieve_datasethandles_or_versionnumbers_of_allversions(
            drs_id=args['drs_id'],
        )
        return result_dict

    def retrieve_file_handles_of_same_dataset(self, **args):
        mandatory_args = ['drs_id', 'data_node', 'version_number']
        esgfpid.utils.check_presence_of_mandatory_args(args, mandatory_args)
        esgfpid.utils.check_noneness_of_mandatory_args(args, mandatory_args)

        result_dict = self.__solr_sender.retrieve_file_handles_of_same_dataset(
            drs_id=args['drs_id'],
            data_node=args['data_node'],
            version_number=args['data_node']
        )
        return result_dict

    def is_solr_switched_off(self):
        return self.__solr_sender.is_switched_off()


import json
import logging
import mock
import sys
sys.path.append('..')
import esgfpid

class SimpleMockRabbitSender(object):

    def __init__(self):
        self.routing_key = None
        self.received_messages = []

    def __remove_date(self, msg):
        if 'message_timestamp' in msg:
            msg['message_timestamp'] = 'anydate'
        return msg

    def send_message_to_queue(self, message):
        self.received_messages.append(self.__remove_date(message))

    def open_rabbit_connection(self):
        pass

    def close_rabbit_connection(self):
        pass

    def finish(self):
        pass

    def start(self):
        pass

    def force_finish(self):
        pass

class MockRabbitSender(object):

    def __init__(self):
        self.routing_key = None
        self.received_messages = []
        self.count = 0

    def __remove_date(self, msg):
        msg['message_timestamp'] = 'anydate'
        return msg

    def send_message_to_queue(self, **args):
        mandatory_args = ['routing_key', 'message']
        esgfpid.utils.check_presence_of_mandatory_args(args, mandatory_args)
        esgfpid.utils.check_noneness_of_mandatory_args(args, mandatory_args)
        self.routing_key = args['routing_key']
        msg = args['message']
        self.received_messages.append(self.__remove_date(msg))

        # Retrieve testcase identifier and test message identifier
        #testcase_identifier = '(no testcase identifier)'
        #try:
        #    testcase_identifier = json.loads(msg)['testcase_identifier']
        #except (TypeError):
        #    testcase_identifier = msg['testcase_identifier']            
        # 
        #testmessage_identifier = '(no testmessage identifier)'
        #try:
        #    testmessage_identifier = json.loads(msg)['testmessage_identifier']
        #except (TypeError):
        #    testmessage_identifier = msg['testmessage_identifier']

        self.count = self.count + 1
        self.received_messages.append(msg)
        #if testcase_identifier is None:
        #    raise ValueError('None1!')
        #if testcase_identifier is None:
        #    raise ValueError('None2!')
        #if testcase_identifier is None:
        #    raise ValueError('None3!')
        #messagelogger.info(testcase_identifier+'#'+testmessage_identifier+'#'+json.dumps(msg))

    def open_rabbit_connection(self):
        pass

    def close_rabbit_connection(self):
        pass



import unittest
import mock
import logging
import copy
import os
import esgfpid.rabbit.asynchronous.thread_feeder
from esgfpid.rabbit.asynchronous.exceptions import OperationNotAllowed

LOGGER = logging.getLogger(__name__)
LOGGER.addHandler(logging.NullHandler())

UNCONFIRMED_TAGS = [1,2,3,4]
UNCONFIRMED_MESSAGES = {'1':'foo1', '2':'foo2', '3':'foo3', '4':'foo4'}


class ThreadFeederTestCase(unittest.TestCase):

    def setUp(self):
        LOGGER.info('######## Next test ########## (%s)' % os.path.basename(__file__))

    def tearDown(self):
        LOGGER.info('#############################')

    def make_feeder(self):

        # Mocked:
        self.thread = mock.MagicMock()
        self.feeder = mock.MagicMock()
        self.statemachine = esgfpid.rabbit.asynchronous.thread_statemachine.StateMachine()
        self.confirmer = esgfpid.rabbit.asynchronous.thread_confirmer.ConfirmReactor()
        feeder = esgfpid.rabbit.asynchronous.thread_feeder.RabbitFeeder(
            self.thread,
            self.statemachine,
            self.confirmer,
            'foo-exchange')
        self.statemachine._StateMachine__state = self.statemachine._StateMachine__IS_AVAILABLE
        return feeder

    def add_leftovers_to_confirmer(self):
        self.confirmer._ConfirmReactor__unconfirmed_delivery_tags = copy.copy(UNCONFIRMED_TAGS)
        self.confirmer._ConfirmReactor__unconfirmed_messages_dict = copy.copy(UNCONFIRMED_MESSAGES)

    # Tests

    #
    # Publish a message
    #

    def test_publish_message_ok(self):

        # Preparation:
        feeder = self.make_feeder()
        feeder._RabbitFeeder__unpublished_messages_queue.put('foo')

        # Run code to be tested:
        num1 = feeder.get_num_unpublished()
        feeder.publish_message()
        num2 = feeder.get_num_unpublished()

        # Check result
        self.assertEquals(num1, 1)
        self.assertEquals(num2, 0)
        self.assertEquals(feeder._RabbitFeeder__message_number, 1)


    def test_publish_message_queue_empty(self):

        # Preparation:
        feeder = self.make_feeder()

        # Run code to be tested:
        num1 = feeder.get_num_unpublished()
        feeder.publish_message()
        num2 = feeder.get_num_unpublished()

        # Check result
        self.assertEquals(num1, 0)
        self.assertEquals(num2, 0)
        self.assertEquals(feeder._RabbitFeeder__message_number, 0)


    def test_publish_while_wanting_to_stop_ok(self):

        # Preparation:
        feeder = self.make_feeder()
        feeder._RabbitFeeder__unpublished_messages_queue.put('foo')
        feeder.statemachine.set_to_wanting_to_stop()

        # Run code to be tested:
        num1 = feeder.get_num_unpublished()
        feeder.publish_message()
        num2 = feeder.get_num_unpublished()

        # Check result
        self.assertEquals(num1, 1)
        self.assertEquals(num2, 0)
        self.assertEquals(feeder._RabbitFeeder__message_number, 1)

    def test_publish_failed(self):

        # Preparation:
        feeder = self.make_feeder()
        feeder._RabbitFeeder__unpublished_messages_queue.put('foo')
        feeder.thread._channel.basic_publish.side_effect = AttributeError

        # Run code to be tested:
        num1 = feeder.get_num_unpublished()
        feeder.publish_message()
        num2 = feeder.get_num_unpublished()

        # Check result
        self.assertEquals(num1, 1)
        self.assertEquals(num2, 1) # None was published, so it stays 1
        self.assertEquals(feeder._RabbitFeeder__message_number, 0)


    def test_wrong_state_1a(self):

        # Preparation:
        feeder = self.make_feeder()
        feeder._RabbitFeeder__unpublished_messages_queue.put('foo')
        feeder.statemachine.set_to_permanently_unavailable()
        feeder.statemachine.could_not_connect = True

        # Run code to be tested:
        num1 = feeder.get_num_unpublished()
        feeder.publish_message()
        num2 = feeder.get_num_unpublished()

        # Check result
        self.assertEquals(num1, 1)
        self.assertEquals(num2, 1) # None was published, so it stays 1
        self.assertEquals(feeder._RabbitFeeder__message_number, 0)

    def test_wrong_state_1b(self):

        # Preparation:
        feeder = self.make_feeder()
        feeder._RabbitFeeder__unpublished_messages_queue.put('foo')
        feeder.statemachine.set_to_permanently_unavailable()
        feeder.statemachine.could_not_connect = False
        feeder.statemachine.closed_by_publisher = True

        # Run code to be tested:
        num1 = feeder.get_num_unpublished()
        feeder.publish_message()
        num2 = feeder.get_num_unpublished()

        # Check result
        self.assertEquals(num1, 1)
        self.assertEquals(num2, 1) # None was published, so it stays 1
        self.assertEquals(feeder._RabbitFeeder__message_number, 0)

    def test_wrong_state_2(self):

        # Preparation:
        feeder = self.make_feeder()
        feeder._RabbitFeeder__unpublished_messages_queue.put('foo')
        feeder.statemachine.set_to_waiting_to_be_available()

        # Run code to be tested:
        num1 = feeder.get_num_unpublished()
        feeder.publish_message()
        num2 = feeder.get_num_unpublished()

        # Check result
        self.assertEquals(num1, 1)
        self.assertEquals(num2, 1) # None was published, so it stays 1
        self.assertEquals(feeder._RabbitFeeder__message_number, 0)

    def test_wrong_state_3(self):

        # Preparation:
        feeder = self.make_feeder()
        feeder._RabbitFeeder__unpublished_messages_queue.put('foo')
        feeder.statemachine._StateMachine__state = self.statemachine._StateMachine__NOT_STARTED_YET

        # Run code to be tested:
        num1 = feeder.get_num_unpublished()
        feeder.publish_message()
        num2 = feeder.get_num_unpublished()

        # Check result
        self.assertEquals(num1, 1)
        self.assertEquals(num2, 1) # None was published, so it stays 1
        self.assertEquals(feeder._RabbitFeeder__message_number, 0)


    #
    # Putter
    #

    def test_put_message_into_queue_of_unsent_messages(self):

        # Preparation:
        feeder = self.make_feeder()

        # Run code to be tested:
        num1 = feeder.get_num_unpublished()
        feeder.put_message_into_queue_of_unsent_messages('foo')
        feeder.put_message_into_queue_of_unsent_messages('bar')
        feeder.put_message_into_queue_of_unsent_messages('baz')
        num2 = feeder.get_num_unpublished()

        # Check result
        self.assertEquals(num1, 0)
        self.assertEquals(num2, 3)
        received = feeder.get_unpublished_messages_as_list_copy()
        self.assertIn('foo', received)
        self.assertIn('bar', received)
        self.assertIn('baz', received)

    #
    # Getter
    #

    def test_get_unpublished_messages_as_list_copy(self):

        # Preparation:
        feeder = self.make_feeder()
        feeder._RabbitFeeder__unpublished_messages_queue.put('foo')
        feeder._RabbitFeeder__unpublished_messages_queue.put('bar')
        feeder._RabbitFeeder__unpublished_messages_queue.put('baz')

        # Run code to be tested:
        mylist = feeder.get_unpublished_messages_as_list_copy()

        # Check result
        self.assertIn('foo', mylist)
        self.assertIn('bar', mylist)
        self.assertIn('baz', mylist)




import esgfpid.defaults
import pika

def get_connection_parameters(credentials, host): # TODO: THis is static. Store as attribute?
    '''
    https://pika.readthedocs.org/en/0.9.6/connecting.html
    class pika.connection.ConnectionParameters(
        host=None, port=None, virtual_host=None, credentials=None, channel_max=None,
        frame_max=None, heartbeat_interval=None, ssl=None, ssl_options=None,
        connection_attempts=None, retry_delay=None, socket_timeout=None, locale=None,
        backpressure_detection=None)
    '''

    # Defaults:
    socket_timeout = esgfpid.defaults.RABBIT_ASYN_SOCKET_TIMEOUT
    connection_attempts = esgfpid.defaults.RABBIT_ASYN_CONNECTION_ATTEMPTS
    retry_delay = esgfpid.defaults.RABBIT_ASYN_CONNECTION_RETRY_DELAY_SECONDS

    # https://pika.readthedocs.org/en/0.9.6/connecting.html
    return pika.ConnectionParameters(
        host=host, # TODO: PORTS ETC.
        credentials=credentials,
        socket_timeout=socket_timeout,
        connection_attempts=connection_attempts,
        retry_delay=retry_delay
    )


def get_credentials(username, password):
    return pika.PlainCredentials(
        username,
        password
    )

def get_properties_for_message_publications():
    properties = pika.BasicProperties(
        delivery_mode=esgfpid.defaults.RABBIT_DELIVERY_MODE,
        content_type='application/json',
    )
    return properties


import logging
import datetime
import json
import pika
import time
import random
import esgfpid.utils
import esgfpid.defaults
import esgfpid.rabbit.synchronous
import esgfpid.rabbit.asynchronous
import esgfpid.rabbit.rabbitutils
import esgfpid.rabbit.nodemanager
from esgfpid.utils import logwarn

# Normal logger:
LOGGER = logging.getLogger(__name__)
LOGGER.addHandler(logging.NullHandler())

ASYNCHRONOUS = esgfpid.defaults.RABBIT_IS_ASYNCHRONOUS

class RabbitMessageSender(object):

    def __init__(self, **args):
        LOGGER.debug('Initializing RabbitMessageSender.')

        mandatory_args = [
            'exchange_name',
            'credentials',
            'test_publication'
        ]
        esgfpid.utils.check_presence_of_mandatory_args(args, mandatory_args)

        self.__node_manager = self.__make_rabbit_settings(args)
        self.__test_publication = args['test_publication']
        self.__server_connector = self.__init_server_connector(args, self.__node_manager)


    def __init_server_connector(self, args, node_manager):
        if ASYNCHRONOUS:
            return esgfpid.rabbit.asynchronous.AsynchronousRabbitConnector(node_manager)
        else:
            logerror(LOGGER, 'Synchronous communication with RabbitMQ is not supported anymore.')
            raise ValueError('Synchronous communication with RabbitMQ is not supported anymore.')
            # The synchronous module has to be fixed. The way of passing credentials was modified
            # and this was not modified in the synchronous module.
            #return esgfpid.rabbit.synchronous.SynchronousServerConnector(**args)

    def open_rabbit_connection(self):
        if not ASYNCHRONOUS:
            return self.__server_connector.open_rabbit_connection()

    def close_rabbit_connection(self):
        if not ASYNCHRONOUS:
            return self.__server_connector.close_rabbit_connection()

    def finish(self):
        if ASYNCHRONOUS:
            self.__server_connector.finish_rabbit_thread()

    def is_finished(self):
        if ASYNCHRONOUS:
            return self.__server_connector.is_finished()
        else:
            return None

    def start(self):
        if ASYNCHRONOUS:
            self.__server_connector.start_rabbit_thread()

    def force_finish(self):
        if ASYNCHRONOUS:
            self.__server_connector.force_finish_rabbit_thread()

    def any_leftovers(self):
        if ASYNCHRONOUS:
            return self.__server_connector.any_leftovers()

    def get_leftovers(self):
        if ASYNCHRONOUS:
            return self.__server_connector.get_leftovers()

    def send_message_to_queue(self, message):
        if self.__test_publication == True:
            message['test_publication'] = True
        return self.__server_connector.send_message_to_queue(message)

    def __make_rabbit_settings(self, args):

        node_manager = esgfpid.rabbit.nodemanager.NodeManager()

        # Add all RabbitMQ nodes:
        for cred in args['credentials']:

            # Open node:
            if cred['password'] == 'jzlnL78ZpExV#_QHz':
                node_manager.add_open_node(
                    username=cred['user'],
                    password='U6-Lke39mN',
                    host=cred['url'],
                    exchange_name=args['exchange_name']
                )

            # Trusted node:
            else:
                node_manager.add_trusted_node(
                    username=cred['user'],
                    password=cred['password'],
                    host=cred['url'],
                    exchange_name=args['exchange_name']
                )

        return node_manager


    def __get_urls_as_list(self, urls_open):

        if isinstance(urls_open, basestring):
            return [urls_open]

        elif urls_open is None:
            logwarn(LOGGER, 'No fallback RabbitMQ URLs specified!')
            return []

        elif not isinstance(urls_open, (list, tuple)):
            msg = ('RabbitMQ URLs are neither list nor string, but %s: %s' %
               (type(args['urls_fallback']), args['urls_fallback']))
            raise ValueError(msg)
        else:
            return urls_open



import unittest
import mock
import logging
import json
import sys
import os
sys.path.append("..")
import esgfpid.rabbit

LOGGER = logging.getLogger(__name__)
LOGGER.addHandler(logging.NullHandler())

class RabbitTestCase(unittest.TestCase):

    def setUp(self):
        LOGGER.info('######## Next test ########## (%s)' % os.path.basename(__file__))

    def tearDown(self):
        LOGGER.info('#############################')

    def __get_args_dict(self):
        return dict(
            urls_fallback='www.rabbit.foo',
            url_preferred=None,
            exchange_name='exch',
            username='rogerRabbit',
            password='mySecretCarrotDream',
            test_publication=False
        )

    def __get_testrabbit_asyn(self):
        args = self.__get_args_dict()
        return esgfpid.rabbit.rabbit.RabbitMessageSender(**args)

    # Tests

    #
    # Asynchronous:
    #

    def test_constructor_mandatory_args(self):

        # Test variables:
        args = self.__get_args_dict()

        # Run code to be tested:
        myrabbit = esgfpid.rabbit.rabbit.RabbitMessageSender(**args)

        # Check result
        self.assertIsInstance(myrabbit, esgfpid.rabbit.rabbit.RabbitMessageSender)
        serverconn = myrabbit._RabbitMessageSender__server_connector
        self.assertIsInstance(serverconn, esgfpid.rabbit.asynchronous.asynchronous.AsynchronousRabbitConnector,
            'Wrong type: %s' % type(serverconn))

    def test_constructor_adapt_args(self):
        ''' Were the args adapted correctly before being passed on
        to the connector? '''

        # Test variables:
        args = self.__get_args_dict()

        # Run code to be tested:
        myrabbit = esgfpid.rabbit.rabbit.RabbitMessageSender(**args)

        # Check wether arguments were correctly changed:
        serverconn = myrabbit._RabbitMessageSender__server_connector
        received_args = serverconn._AsynchronousRabbitConnector__args
        self.assertEquals(received_args[ 'exchange_name'], 'exch')
        self.assertEquals(received_args['url_preferred'], 'www.rabbit.foo') # the only fallback url becomes the preferred
        self.assertEquals(received_args['urls_fallback'], []) # no fallback url is left
        self.assertIn('credentials', received_args)
        self.assertNotIn('username', received_args)
        self.assertNotIn('password', received_args)

    def test_finish_ok(self):
        '''
        If we call finish() on the sender, it calls finish_rabbit_thread() on
        the connector, which calls finish_rabbit_thread() on the thread controller.
        '''

        # Prepare
        args = self.__get_args_dict()
        testrabbit = esgfpid.rabbit.rabbit.RabbitMessageSender(**args)

        # Replace the connector with a mock:
        testrabbit._RabbitMessageSender__server_connector = mock.MagicMock()
        mock_connector = testrabbit._RabbitMessageSender__server_connector
        
        # Run code to be tested
        testrabbit.finish()

        # Check result:
        mock_connector.finish_rabbit_thread.assert_called_with()

    def test_force_finish_ok(self):
        '''
        If we call force_finish() on the sender, it calls force_finish_rabbit_thread() on
        the connector, which calls force_finish_rabbit_thread() on the thread controller.
        '''

        # Prepare
        args = self.__get_args_dict()
        testrabbit = esgfpid.rabbit.rabbit.RabbitMessageSender(**args)

        # Replace the connector with a mock:
        testrabbit._RabbitMessageSender__server_connector = mock.MagicMock()
        mock_connector = testrabbit._RabbitMessageSender__server_connector
        
        # Run code to be tested
        testrabbit.force_finish()

        # Check result:
        mock_connector.force_finish_rabbit_thread.assert_called_with()

    def test_start_ok(self):
        '''
        If we call start() on the sender, it calls start_rabbit_thread() on
        the connector, which calls start_rabbit_thread(**args) on the thread controller.
        '''

        # Prepare
        args = self.__get_args_dict()
        testrabbit = esgfpid.rabbit.rabbit.RabbitMessageSender(**args)

        # Replace the connector with a mock:
        testrabbit._RabbitMessageSender__server_connector = mock.MagicMock()
        mock_connector = testrabbit._RabbitMessageSender__server_connector
        
        # Run code to be tested
        testrabbit.start()

        # Check result:
        mock_connector.start_rabbit_thread.assert_called_with()


    def test_is_finished_ok(self):
        '''
        If we call is_finished() on the sender, it calls is_finished() on
        the connector, which calls is_finished() on the thread controller.
        '''

        # Prepare
        args = self.__get_args_dict()
        testrabbit = esgfpid.rabbit.rabbit.RabbitMessageSender(**args)

        # Replace the connector with a mock:
        testrabbit._RabbitMessageSender__server_connector = mock.MagicMock()
        mock_connector = testrabbit._RabbitMessageSender__server_connector
                
        # Run code to be tested
        res = testrabbit.is_finished()

        # Check result:
        self.assertTrue(res, 'We asked the mock to return True, not %s.' % res)
        mock_connector.is_finished.assert_called_with()

    def test_leftovers_ok(self):
        '''
        If we call get_leftovers() on the sender, it calls get_leftovers() on
        the connector, which calls get_leftovers() on the thread controller.
        '''
        # Prepare
        args = self.__get_args_dict()
        testrabbit = esgfpid.rabbit.rabbit.RabbitMessageSender(**args)

        # Replace the connector with a mock:
        testrabbit._RabbitMessageSender__server_connector = mock.MagicMock()
        mock_connector = testrabbit._RabbitMessageSender__server_connector
        mock_connector.any_leftovers.return_value = True
        mock_connector.get_leftovers.return_value = ['a', 'b', 'c']
        
        # Run code to be tested
        result_bool = testrabbit.any_leftovers()
        result_list = testrabbit.get_leftovers()

        # Check result:
        self.assertTrue(result_bool,
            'We asked the mock to return True, not %s.' % result_bool)
        self.assertEquals(result_list, ['a', 'b', 'c'],
            'We asked the mock to return a list, not %s.' % result_list)
        mock_connector.any_leftovers.assert_called_with()
        mock_connector.get_leftovers.assert_called_with()

    def test_send_message_to_queue(self):

        # Prepare
        msg = 'FOOBAR'
        args = self.__get_args_dict()
        testrabbit = esgfpid.rabbit.rabbit.RabbitMessageSender(**args)

        # Replace the connector with a mock:
        testrabbit._RabbitMessageSender__server_connector = mock.MagicMock()
        mock_connector = testrabbit._RabbitMessageSender__server_connector
                
        # Run code to be tested
        testrabbit.send_message_to_queue(msg)

        # Check result
        mock_connector.send_message_to_queue.assert_called_with(msg)



import unittest
import mock
import logging
import json
import sys
import esgfpid.assistant.publish
import tests.mocks.responsemock
import tests.mocks.solrmock
import tests.mocks.rabbitmock
import tests.utils as utils
from tests.utils import compare_json_return_errormessage as error_message
from esgfpid.defaults import ROUTING_KEY_BASIS as ROUTING_KEY_BASIS


# Logging
LOGGER = logging.getLogger(__name__)
LOGGER.addHandler(logging.NullHandler())

# Test resources:
from resources.TESTVALUES import TESTVALUES as TESTVALUES


class PublishTestCase(unittest.TestCase):

    def setUp(self):
        LOGGER.info('######## Next test (%s) ##########', __name__)

    def tearDown(self):
        LOGGER.info('#############################')

    def __get_normal_init_args_dataset(self, testcoupler):
        return dict(
            prefix=TESTVALUES['prefix'],
            drs_id=TESTVALUES['drs_id1'],
            version_number=TESTVALUES['version_number1'],
            is_replica=False,
            thredds_service_path=TESTVALUES['thredds_service_path'],
            data_node=TESTVALUES['data_node'],
            coupler=testcoupler,
            consumer_solr_url='does_not_matter'
        )

    def __get_normal_rabbit_task_dataset(self):
        filehandle = 'hdl:'+TESTVALUES['prefix']+'/'+TESTVALUES['file_suffix1']
        return {
            "handle": "hdl:"+TESTVALUES['prefix']+'/'+TESTVALUES['suffix1'],
            "aggregation_level": "dataset",
            "operation": "publish",
            "message_timestamp":"anydate",
            "drs_id":TESTVALUES['drs_id1'],
            "version_number":TESTVALUES['version_number1'],
            "files":[filehandle],
            "is_replica":False,
            "data_node":TESTVALUES['data_node'],
            "ROUTING_KEY": ROUTING_KEY_BASIS+'publication.dataset.orig',
            "consumer_solr_url": "does_not_matter"
        }

    def __get_normal_file_args(self):
        return dict(
            file_name = TESTVALUES['file_name1'],
            file_handle = 'hdl:'+TESTVALUES['prefix']+'/'+TESTVALUES['file_suffix1'],
            checksum = TESTVALUES['checksum1'],
            file_size = TESTVALUES['file_size1'],
            publish_path = TESTVALUES['publish_path1'],
            prefix = TESTVALUES['prefix'],
            checksum_type=TESTVALUES['checksum_type1'],
            file_version=TESTVALUES['file_version1'] # string
        )

    def __get_normal_rabbit_task_file(self):
        expected_data_url = ('http://'+TESTVALUES['data_node']+
            '/'+TESTVALUES['thredds_service_path']+
            '/'+TESTVALUES['publish_path1'])
        expected_rabbit_task = {
            "handle": "hdl:"+TESTVALUES['prefix']+'/'+TESTVALUES['file_suffix1'],
            "aggregation_level": "file",
            "operation": "publish",
            "message_timestamp":"anydate",
            "is_replica":False,
            "file_name": TESTVALUES['file_name1'],
            "data_url":expected_data_url,
            "data_node":TESTVALUES['data_node'],
            "file_size":TESTVALUES['file_size1'],
            "checksum":TESTVALUES['checksum1'],
            "checksum_type":TESTVALUES['checksum_type1'],
            "file_version":TESTVALUES['file_version1'],
            "parent_dataset":"hdl:"+TESTVALUES['prefix']+'/'+TESTVALUES['suffix1'],
            "ROUTING_KEY": ROUTING_KEY_BASIS+'publication.file.orig',
        }
        return expected_rabbit_task

    def __make_patched_testcoupler(self, solr_off=False):
        testcoupler = esgfpid.coupling.Coupler(
            handle_prefix = TESTVALUES['prefix'],
            messaging_service_urls_open = 'rabbit_should_not_be_used',
            messaging_service_url_trusted = TESTVALUES['url_rabbit_trusted'],
            messaging_service_exchange_name = TESTVALUES['rabbit_exchange_name'],
            messaging_service_username_open = TESTVALUES['rabbit_username_open'],
            messaging_service_username_trusted = TESTVALUES['rabbit_username_trusted'],
            data_node = TESTVALUES['data_node'],
            thredds_service_path = TESTVALUES['thredds_service_path'],
            solr_url = 'solr_should_not_be_used',
            solr_https_verify=True,
            solr_switched_off=solr_off
        )
        # Replace objects that interact with servers with mocks
        # Patch rabbit
        rabbitmock = tests.mocks.rabbitmock.SimpleMockRabbitSender()
        testcoupler._Coupler__rabbit_message_sender = rabbitmock
        # Solr has to be atched for each test case!
        return testcoupler

    def __patch_coupler_with_solr_mock(self, testcoupler, previous_files, raise_error=None):
        if raise_error is None:
            solrmock = tests.mocks.solrmock.MockSolrInteractor(previous_files=previous_files)
        else:
            solrmock = tests.mocks.solrmock.MockSolrInteractor(previous_files=previous_files, raise_error=raise_error)
        testcoupler._Coupler__solr_sender = solrmock

    def __get_received_message_from_rabbit_mock(self, coupler, index):
        return coupler._Coupler__rabbit_message_sender.received_messages[index]

    def __make_test_assistant(self, testcoupler=None):
        if testcoupler is None:
            testcoupler = self.__make_patched_testcoupler()
        args = self.__get_normal_init_args_dataset(testcoupler)
        assistant = esgfpid.assistant.publish.DatasetPublicationAssistant(**args)
        return assistant

    ### Actual test cases: ###

    #
    # Testing only init (passing info about dataset, but not about file yet):
    #

    def test_version_number_is_string_ok(self):

        # Preparations:
        testcoupler = self.__make_patched_testcoupler()
        self.__patch_coupler_with_solr_mock(testcoupler, None) # Solr returns []
        args = self.__get_normal_init_args_dataset(testcoupler)
        args['is_replica'] = False
        args['version_number'] = str(args['version_number'])

        # Run code to be tested:
        assistant = esgfpid.assistant.publish.DatasetPublicationAssistant(**args)

        # Check result:
        self.assertIsInstance(assistant, esgfpid.assistant.publish.DatasetPublicationAssistant,
            'Constructor failed.')

    def test_version_number_has_characters_error(self):

        # Preparations:
        testcoupler = self.__make_patched_testcoupler()
        self.__patch_coupler_with_solr_mock(testcoupler, None) # Solr returns []
        args = self.__get_normal_init_args_dataset(testcoupler)
        args['is_replica'] = False
        args['version_number'] = 'abcdef'

        # Run code to be tested and check exception:
        with self.assertRaises(esgfpid.exceptions.ArgumentError):
            assistant = esgfpid.assistant.publish.DatasetPublicationAssistant(**args)

    def test_init_normal(self):

        # Preparations:
        testcoupler = self.__make_patched_testcoupler()
        self.__patch_coupler_with_solr_mock(testcoupler, None) # Solr returns []
        args = self.__get_normal_init_args_dataset(testcoupler)
        args['is_replica'] = False

        # Run code to be tested:
        assistant = esgfpid.assistant.publish.DatasetPublicationAssistant(**args)

        # Check result:
        self.assertIsInstance(assistant, esgfpid.assistant.publish.DatasetPublicationAssistant,
            'Constructor failed.')

    def test_init_no_consumer_solr_url(self):

        # Preparations:
        testcoupler = self.__make_patched_testcoupler()
        self.__patch_coupler_with_solr_mock(testcoupler, None) # Solr returns []
        args = self.__get_normal_init_args_dataset(testcoupler)
        args['is_replica'] = False
        del args['consumer_solr_url']

        # Run code to be tested and check exception:
        with self.assertRaises(esgfpid.exceptions.ArgumentError):
            assistant = esgfpid.assistant.publish.DatasetPublicationAssistant(**args)

    def test_init_consumer_solr_url_is_none(self):

        # Preparations:
        testcoupler = self.__make_patched_testcoupler()
        self.__patch_coupler_with_solr_mock(testcoupler, None) # Solr returns []
        args = self.__get_normal_init_args_dataset(testcoupler)
        args['is_replica'] = False
        args['consumer_solr_url'] = None

        # Run code to be tested:
        assistant = esgfpid.assistant.publish.DatasetPublicationAssistant(**args)

        # Check result:
        self.assertIsInstance(assistant, esgfpid.assistant.publish.DatasetPublicationAssistant,
            'Constructor failed.')

    def test_init_string_replica_flag_ok(self):

        # Preparations:
        testcoupler = self.__make_patched_testcoupler()
        self.__patch_coupler_with_solr_mock(testcoupler, None) # Solr returns []
        args = self.__get_normal_init_args_dataset(testcoupler)
        args['is_replica'] = 'False'

        # Run code to be tested:
        assistant = esgfpid.assistant.publish.DatasetPublicationAssistant(**args)

        # Check result:
        self.assertIsInstance(assistant, esgfpid.assistant.publish.DatasetPublicationAssistant,
            'Constructor failed.')
        self.assertFalse(assistant._DatasetPublicationAssistant__is_replica, 'Replica flag should be False if it is passed as "False".')
    
    def test_init_string_wrong_replica_flag(self):

        # Preparations:
        testcoupler = self.__make_patched_testcoupler()
        self.__patch_coupler_with_solr_mock(testcoupler, None) # Solr returns []
        args = self.__get_normal_init_args_dataset(testcoupler)
        args['is_replica'] = 'Maybe'

        # Run code to be tested:
        with self.assertRaises(esgfpid.exceptions.ArgumentError) as raised:
            assistant = esgfpid.assistant.publish.DatasetPublicationAssistant(**args)
        self.assertIn('"Maybe" could not be parsed to boolean', raised.exception.message,
            'Unexpected error message: %s' % raised.exception.message)

    #
    # Testing entire publication
    # First publication, no previous info on solr, no consistency check
    #

    def test_normal_publication_ok(self):

        # Preparations:
        testcoupler = self.__make_patched_testcoupler()
        self.__patch_coupler_with_solr_mock(testcoupler, None) # Solr returns []
        assistant = self.__make_test_assistant(testcoupler)
        args = self.__get_normal_file_args()

        # Run code to be tested:
        assistant.add_file(**args)
        assistant.dataset_publication_finished()

        # Check result (dataset):
        received_rabbit_task = self.__get_received_message_from_rabbit_mock(testcoupler, 0)
        expected_rabbit_task = self.__get_normal_rabbit_task_dataset()
        same = utils.is_json_same(expected_rabbit_task, received_rabbit_task)
        self.assertTrue(same, error_message(expected_rabbit_task, received_rabbit_task))

        # Check result (file):
        received_rabbit_task = self.__get_received_message_from_rabbit_mock(testcoupler, 1)
        expected_rabbit_task = self.__get_normal_rabbit_task_file()
        same = utils.is_json_same(expected_rabbit_task, received_rabbit_task)
        self.assertTrue(same, error_message(expected_rabbit_task, received_rabbit_task))

    def test_normal_publication_sev_files_ok(self):

        # Test variables
        prefix = TESTVALUES['prefix']
        handle1 = 'hdl:'+prefix+'/456'
        handle2 = 'hdl:'+prefix+'/789'

        # Preparations:
        testcoupler = self.__make_patched_testcoupler()
        self.__patch_coupler_with_solr_mock(testcoupler, None) # Solr returns []
        assistant = self.__make_test_assistant(testcoupler)
        args1 = self.__get_normal_file_args()
        args1['file_handle'] = handle1
        args2 = self.__get_normal_file_args()
        args2['file_handle'] = handle2

        # Run code to be tested:
        assistant.add_file(**args1)
        assistant.add_file(**args2)
        assistant.dataset_publication_finished()

        # Check result (dataset):
        received_rabbit_task = self.__get_received_message_from_rabbit_mock(testcoupler, 0)
        expected_rabbit_task = self.__get_normal_rabbit_task_dataset()
        expected_rabbit_task['files'] = [handle2, handle1]
        same = utils.is_json_same(expected_rabbit_task, received_rabbit_task)
        self.assertTrue(same, error_message(expected_rabbit_task, received_rabbit_task))

        # Check result (file):
        received_rabbit_task = self.__get_received_message_from_rabbit_mock(testcoupler, 1)
        expected_rabbit_task = self.__get_normal_rabbit_task_file()
        expected_rabbit_task['handle'] = handle1
        same = utils.is_json_same(expected_rabbit_task, received_rabbit_task)
        self.assertTrue(same, error_message(expected_rabbit_task, received_rabbit_task))

        # Check result (file):
        received_rabbit_task = self.__get_received_message_from_rabbit_mock(testcoupler, 2)
        expected_rabbit_task = self.__get_normal_rabbit_task_file()
        expected_rabbit_task['handle'] = handle2
        same = utils.is_json_same(expected_rabbit_task, received_rabbit_task)
        self.assertTrue(same, error_message(expected_rabbit_task, received_rabbit_task))

    def test_add_file_wrong_prefix(self):

        # Preparations:
        testcoupler = self.__make_patched_testcoupler()
        self.__patch_coupler_with_solr_mock(testcoupler, None) # Solr returns []
        assistant = self.__make_test_assistant(testcoupler)
        args = self.__get_normal_file_args()
        args['file_handle'] = 'hdl:1234/56789'

        # Run code to be tested:
        with self.assertRaises(esgfpid.exceptions.ESGFException) as raised:
            assistant.add_file(**args)
        self.assertIn('ESGF rule violation', raised.exception.message,
            'Unexpected message: %s' % raised.exception.message)

    def test_add_file_without_hdl_in_handle(self):

        # Preparations:
        testcoupler = self.__make_patched_testcoupler()
        self.__patch_coupler_with_solr_mock(testcoupler, None) # Solr returns []
        assistant = self.__make_test_assistant(testcoupler)
        args = self.__get_normal_file_args()
        handle = TESTVALUES['prefix']+'/'+TESTVALUES['file_suffix1']
        args['file_handle'] = handle

        # Run code to be tested:
        assistant.add_file(**args)
        assistant.dataset_publication_finished()

        # Check result (dataset):
        received_rabbit_task = self.__get_received_message_from_rabbit_mock(testcoupler, 0)
        expected_rabbit_task = self.__get_normal_rabbit_task_dataset()
        same = utils.is_json_same(expected_rabbit_task, received_rabbit_task)
        self.assertTrue(same, error_message(expected_rabbit_task, received_rabbit_task))

        # Check result (file):
        received_rabbit_task = self.__get_received_message_from_rabbit_mock(testcoupler, 1)
        expected_rabbit_task = self.__get_normal_rabbit_task_file()
        same = utils.is_json_same(expected_rabbit_task, received_rabbit_task)
        self.assertTrue(same, error_message(expected_rabbit_task, received_rabbit_task))

    def test_add_file_fileversion_as_integer_filesize_as_string(self):

        # Preparations:
        testcoupler = self.__make_patched_testcoupler()
        assistant = self.__make_test_assistant(testcoupler)
        self.__patch_coupler_with_solr_mock(testcoupler, None) # Solr returns []
        args = self.__get_normal_file_args()
        args['file_version'] = int(args['file_version'])
        args['file_size'] = str(args['file_size'])

        # Run code to be tested:
        assistant.add_file(**args)
        assistant.dataset_publication_finished()

        # Check result (file):
        received_rabbit_task = self.__get_received_message_from_rabbit_mock(testcoupler, 1)
        expected_rabbit_task = self.__get_normal_rabbit_task_file()
        same = utils.is_json_same(expected_rabbit_task, received_rabbit_task)
        self.assertTrue(same, error_message(expected_rabbit_task, received_rabbit_task))

    def test_add_file_filesize_has_characters(self):

        # Preparations:
        testcoupler = self.__make_patched_testcoupler()
        assistant = self.__make_test_assistant(testcoupler)
        self.__patch_coupler_with_solr_mock(testcoupler, None) # Solr returns []
        args = self.__get_normal_file_args()
        args['file_size'] = 'abc'

        # Run code to be tested:
        with self.assertRaises(esgfpid.exceptions.ArgumentError):
            assistant.add_file(**args)
        #assistant.dataset_publication_finished()

        # Check result (file):
        #received_rabbit_task = self.__get_received_message_from_rabbit_mock(testcoupler, 1)
        #expected_rabbit_task = self.__get_normal_rabbit_task_file()
        #same = utils.is_json_same(expected_rabbit_task, received_rabbit_task)
        #self.assertTrue(same, error_message(expected_rabbit_task, received_rabbit_task))

    def test_normal_publication_without_solr_access(self):
        '''
        Solr was switched off and raises an errors, but this is caught.
        The consistency check is not run, but the rest of the publication is the same.
        '''

        # Preparations:  
        testcoupler = self.__make_patched_testcoupler()
        # Patch the solr module with a different mock:
        error_to_be_raised = esgfpid.exceptions.SolrSwitchedOff
        self.__patch_coupler_with_solr_mock(testcoupler, None, error_to_be_raised) # makes mock raise this error
        assistant = self.__make_test_assistant(testcoupler)
        # Other variables:
        args = self.__get_normal_file_args()

        # Run code to be tested:
        assistant.add_file(**args)
        assistant.dataset_publication_finished()

        # Check result (dataset):
        received_rabbit_task = self.__get_received_message_from_rabbit_mock(testcoupler, 0)
        expected_rabbit_task = self.__get_normal_rabbit_task_dataset()
        same = utils.is_json_same(expected_rabbit_task, received_rabbit_task)
        self.assertTrue(same, error_message(expected_rabbit_task, received_rabbit_task))

        # Check result (file):
        received_rabbit_task = self.__get_received_message_from_rabbit_mock(testcoupler, 1)
        expected_rabbit_task = self.__get_normal_rabbit_task_file()
        same = utils.is_json_same(expected_rabbit_task, received_rabbit_task)
        self.assertTrue(same, error_message(expected_rabbit_task, received_rabbit_task))

    def test_normal_publication_solr_returns_none(self):

        # Preparations:  
        testcoupler = self.__make_patched_testcoupler()
        # Patch the solr module with a different mock:
        self.__patch_coupler_with_solr_mock(testcoupler, 'NONE') # makes mock return None - this should never happen in reality!
        assistant = self.__make_test_assistant(testcoupler)
        # Other variables:
        args = self.__get_normal_file_args()

        # Run code to be tested:
        assistant.add_file(**args)
        assistant.dataset_publication_finished()

        # Check result (dataset):
        received_rabbit_task = self.__get_received_message_from_rabbit_mock(testcoupler, 0)
        expected_rabbit_task = self.__get_normal_rabbit_task_dataset()
        same = utils.is_json_same(expected_rabbit_task, received_rabbit_task)
        self.assertTrue(same, error_message(expected_rabbit_task, received_rabbit_task))

    #
    # Testing entire RE-publication
    # Republication, solr provides info on previous files, consistency check
    #

    def test_normal_publication_with_pos_consis_check(self):

        # Test variables:
        args = self.__get_normal_file_args()
        prev = [args['file_handle']]
        # Preparations:  
        testcoupler = self.__make_patched_testcoupler()
        self.__patch_coupler_with_solr_mock(testcoupler, prev) # solr returns file list
        assistant = self.__make_test_assistant(testcoupler)

        # Run code to be tested:
        assistant.add_file(**args)
        assistant.dataset_publication_finished()

        # Check result (dataset):
        received_rabbit_task = self.__get_received_message_from_rabbit_mock(testcoupler, 0)
        expected_rabbit_task = self.__get_normal_rabbit_task_dataset()
        expected_rabbit_task['files'] = [args['file_handle']]
        same = utils.is_json_same(expected_rabbit_task, received_rabbit_task)
        self.assertTrue(same, error_message(expected_rabbit_task, received_rabbit_task))

        # Check result (file):
        received_rabbit_task = self.__get_received_message_from_rabbit_mock(testcoupler, 1)
        expected_rabbit_task = self.__get_normal_rabbit_task_file()
        same = utils.is_json_same(expected_rabbit_task, received_rabbit_task)
        self.assertTrue(same, error_message(expected_rabbit_task, received_rabbit_task))

    def test_normal_publication_with_pos_consis_check_missing_hdl(self):

        # Test variables:
        args = self.__get_normal_file_args()
        handle_no_hdl = TESTVALUES['prefix']+'/'+TESTVALUES['file_suffix1']
        handle_with_hdl = 'hdl:'+handle_no_hdl
        prev = [handle_with_hdl] # solr always adds hdl before returning it!
        args['file_handle'] = handle_no_hdl # Handle without hdl is published!
        # Preparations:  
        testcoupler = self.__make_patched_testcoupler()
        self.__patch_coupler_with_solr_mock(testcoupler, prev) # solr returns file list
        assistant = self.__make_test_assistant(testcoupler)

        # Run code to be tested:
        assistant.add_file(**args)
        assistant.dataset_publication_finished()

        # Check result (dataset):
        received_rabbit_task = self.__get_received_message_from_rabbit_mock(testcoupler, 0)
        expected_rabbit_task = self.__get_normal_rabbit_task_dataset()
        expected_rabbit_task['files'] = ['hdl:'+args['file_handle']]
        same = utils.is_json_same(expected_rabbit_task, received_rabbit_task)
        self.assertTrue(same, error_message(expected_rabbit_task, received_rabbit_task))

        # Check result (file):
        received_rabbit_task = self.__get_received_message_from_rabbit_mock(testcoupler, 1)
        expected_rabbit_task = self.__get_normal_rabbit_task_file()
        same = utils.is_json_same(expected_rabbit_task, received_rabbit_task)
        self.assertTrue(same, error_message(expected_rabbit_task, received_rabbit_task))

    def test_normal_publication_with_neg_consis_check_missing_file(self):

        # Test variables:
        file_handle = TESTVALUES['prefix']+'/'+TESTVALUES['file_suffix1']
        another_file_handle = TESTVALUES['prefix']+'/random_suffix_abc123'
        prev = [file_handle, another_file_handle]
        file_name = TESTVALUES['file_name1']
        checksum = TESTVALUES['checksum1']
        file_size = TESTVALUES['file_size1']
        publish_path = TESTVALUES['publish_path1']
        prefix = TESTVALUES['prefix']
        checksum_type=TESTVALUES['checksum_type1']

        # Preparations:
        testcoupler = self.__make_patched_testcoupler()
        self.__patch_coupler_with_solr_mock(testcoupler, prev) # solr returns file list
        assistant = self.__make_test_assistant(testcoupler)

        # Prepare:
        assistant.add_file(
            file_name=file_name,
            file_handle=file_handle,
            checksum=checksum,
            file_size=file_size,
            publish_path=publish_path,
            checksum_type=checksum_type,
            file_version=1
        )

        # Run code to be tested and check exception:
        with self.assertRaises(esgfpid.exceptions.InconsistentFilesetException):
            assistant.dataset_publication_finished()

    def test_normal_publication_with_neg_consis_check_too_many_files(self):

        # Test variables:
        file_handle = TESTVALUES['prefix']+'/'+TESTVALUES['file_suffix1']
        prev = [file_handle]
        file_name = TESTVALUES['file_name1']
        checksum = TESTVALUES['checksum1']
        file_size = TESTVALUES['file_size1']
        publish_path = TESTVALUES['publish_path1']
        prefix = TESTVALUES['prefix']
        file_handle2 = TESTVALUES['prefix']+'/'+TESTVALUES['file_suffix2']
        checksum_type=TESTVALUES['checksum_type1']

        # Prepare:
        testcoupler = self.__make_patched_testcoupler()
        self.__patch_coupler_with_solr_mock(testcoupler, prev) # solr returns file list
        assistant = self.__make_test_assistant(testcoupler)

        assistant.add_file(
            file_name=file_name,
            file_handle=file_handle,
            checksum=checksum,
            file_size=file_size,
            publish_path=publish_path,
            checksum_type=checksum_type,
            file_version=1
        )
        assistant.add_file(
            file_name=file_name,
            file_handle=file_handle2,
            checksum=checksum,
            file_size=file_size,
            publish_path=publish_path,
            checksum_type=checksum_type,
            file_version=1
        )

        # Run code to be tested and check exception:
        with self.assertRaises(esgfpid.exceptions.InconsistentFilesetException):
            assistant.dataset_publication_finished()

    #
    # Testing invalid operations
    # (publication actions occur at the wrong time / wrong state)
    #

    def test_add_file_too_late(self):

        # Test variables:
        args = self.__get_normal_file_args()

        # Prepare:
        testcoupler = self.__make_patched_testcoupler(solr_off=True)
        assistant = self.__make_test_assistant(testcoupler)
        assistant.add_file(**args)
        assistant.dataset_publication_finished()

        # Run code to be tested and check exception:
        with self.assertRaises(esgfpid.exceptions.OperationUnsupportedException):
            assistant.add_file(**args)

    def test_finish_too_late(self):

        # Test variables:
        args = self.__get_normal_file_args()

        # Preparations
        testcoupler = self.__make_patched_testcoupler(solr_off=True)
        assistant = self.__make_test_assistant(testcoupler)
        assistant.add_file(**args)
        assistant.dataset_publication_finished()

        # Run code to be tested:
        with self.assertRaises(esgfpid.exceptions.OperationUnsupportedException):
            assistant.dataset_publication_finished()

    def test_finish_too_early(self):

        # Test variables:
        testcoupler = self.__make_patched_testcoupler(solr_off=True)
        assistant = self.__make_test_assistant(testcoupler)

        # Run code to be tested:
        with self.assertRaises(esgfpid.exceptions.OperationUnsupportedException):
            assistant.dataset_publication_finished()

    #
    # Testing replica publication
    #

    def test_normal_publication_replica_flag_bool_ok(self):

        # Preparations:
        testcoupler = self.__make_patched_testcoupler(solr_off=True) # solr switched off, no consistency check
        args = self.__get_normal_init_args_dataset(testcoupler)
        args['is_replica'] = True
        assistant = esgfpid.assistant.publish.DatasetPublicationAssistant(**args)
        fileargs = self.__get_normal_file_args()

        # Run code to be tested:
        assistant.add_file(**fileargs)
        assistant.dataset_publication_finished()

        # Check result (dataset):
        received_rabbit_task = self.__get_received_message_from_rabbit_mock(testcoupler, 0)
        expected_rabbit_task = self.__get_normal_rabbit_task_dataset()
        expected_rabbit_task['is_replica'] = True
        expected_rabbit_task['ROUTING_KEY'] = ROUTING_KEY_BASIS+'publication.dataset.replica'
        same = utils.is_json_same(expected_rabbit_task, received_rabbit_task)
        self.assertTrue(same, error_message(expected_rabbit_task, received_rabbit_task))

        # Check result (file):
        received_rabbit_task = self.__get_received_message_from_rabbit_mock(testcoupler, 1)
        expected_rabbit_task = self.__get_normal_rabbit_task_file()
        expected_rabbit_task['is_replica'] = True
        expected_rabbit_task['ROUTING_KEY'] = ROUTING_KEY_BASIS+'publication.file.replica'
        same = utils.is_json_same(expected_rabbit_task, received_rabbit_task)
        self.assertTrue(same, error_message(expected_rabbit_task, received_rabbit_task))

    def test_normal_publication_replica_flag_string_ok(self):

        # Preparations:
        testcoupler = self.__make_patched_testcoupler()
        self.__patch_coupler_with_solr_mock(testcoupler, None) # Solr returns []
        args = self.__get_normal_init_args_dataset(testcoupler)
        args['is_replica'] = 'True'
        assistant = esgfpid.assistant.publish.DatasetPublicationAssistant(**args)
        fileargs = self.__get_normal_file_args()

        # Run code to be tested:
        assistant.add_file(**fileargs)
        assistant.dataset_publication_finished()

        # Check result (dataset):
        received_rabbit_task = self.__get_received_message_from_rabbit_mock(testcoupler, 0)
        expected_rabbit_task = self.__get_normal_rabbit_task_dataset()
        expected_rabbit_task['is_replica'] = True
        expected_rabbit_task['ROUTING_KEY'] = ROUTING_KEY_BASIS+'publication.dataset.replica'
        same = utils.is_json_same(expected_rabbit_task, received_rabbit_task)
        self.assertTrue(same, error_message(expected_rabbit_task, received_rabbit_task))

        # Check result (file):
        received_rabbit_task = self.__get_received_message_from_rabbit_mock(testcoupler, 1)
        expected_rabbit_task = self.__get_normal_rabbit_task_file()
        expected_rabbit_task['is_replica'] = True
        expected_rabbit_task['ROUTING_KEY'] = ROUTING_KEY_BASIS+'publication.file.replica'
        same = utils.is_json_same(expected_rabbit_task, received_rabbit_task)
        self.assertTrue(same, error_message(expected_rabbit_task, received_rabbit_task))

    #
    # Helper tests
    #

    def test_get_dataset_handle(self):

        # Preparations:
        assistant = self.__make_test_assistant()

        # Run code to be tested:
        handle = assistant.get_dataset_handle()

        # Check result:
        exp_handle = 'hdl:'+TESTVALUES['prefix']+'/'+TESTVALUES['suffix1']
        self.assertEqual(handle, exp_handle,
            'Wrong handle returned.')


import unittest
import mock
import logging
import json
import requests
import esgfpid.solr.solr
import tests.mocks.responsemock


# Logging:
LOGGER = logging.getLogger(__name__)
LOGGER.addHandler(logging.NullHandler())

# Test resources:
from tests.resources.IGNORE.TESTVALUES_IGNORE import TESTVALUES_SOLR
SOLR_URL = 'http://'+TESTVALUES_SOLR['solr_url']
PREFIX = TESTVALUES_SOLR['prefix']
DRS_ID = TESTVALUES_SOLR['drs_id']
VERSION_NUMBER =  TESTVALUES_SOLR['version_number']
DATA_NODE = TESTVALUES_SOLR['data_node']
FILES = TESTVALUES_SOLR['files']

class SolrIntegrationTestCase(unittest.TestCase):

    def setUp(self):
        LOGGER.info('######## Next test ##########')

    def tearDown(self):
        LOGGER.info('#############################')

    def make_testsolr_with_access(self):
        testsolr = esgfpid.solr.solr.SolrInteractor(
            solr_url = SOLR_URL,
            prefix = PREFIX
        )
        return testsolr

    # Actual tests:

    def test_send_query_ok(self):
        '''
        Sending basic query that returns number of all files found.'''

        # Test variables
        testsolr = self.make_testsolr_with_access()
        query = testsolr.make_solr_base_query()

        # Run code to be tested:
        response = testsolr.send_query(query)

        # Check result:
        # We expect the mock response to be returned and parsed to JSON.
        self.assertTrue('responseHeader' in response, 'Solr response to query %s had no responseHeader: %s' % (query, response))
        self.assertTrue('facet_counts' in response, 'Solr response to query %s had no facet_counts: %s' % (query, response))
        self.assertTrue('response' in response, 'Solr response to query %s had no response: %s' % (query, response))


    def test_send_query_connection_error(self):

        # Test variables
        testsolr = esgfpid.solr.solr.SolrInteractor(
            solr_url = 'https://ihopethisdoesnotexist.de',
            prefix = PREFIX
        )
        query = 'doesntmatter'

        # Run code to be tested and check exception:
        with self.assertRaises(esgfpid.exceptions.SolrError) as raised:
            response = testsolr.send_query(query)
        self.assertIn('ConnectionError', raised.exception.message)

    def test_send_query_404(self):

        # Test variables
        testsolr = esgfpid.solr.solr.SolrInteractor(
            solr_url = SOLR_URL+'/hahahah/hohohoh/',
            prefix = PREFIX
        )
        query = 'doesntmatter'

        # Run code to be tested and check exception:
        with self.assertRaises(esgfpid.exceptions.SolrError) as raised:
            response = testsolr.send_query(query)
        self.assertIn('HTTP 404', raised.exception.message)


    def test_send_query_400(self):

        # Test variables
        testsolr = esgfpid.solr.solr.SolrInteractor(
            solr_url = SOLR_URL,
            prefix = PREFIX
        )
        query = 'foo'

        # Run code to be tested and check exception:
        with self.assertRaises(esgfpid.exceptions.SolrError) as raised:
            response = testsolr.send_query(query)
        self.assertIn('Solr replied with code 400', raised.exception.message)



import pika
import copy
import logging
import random
import esgfpid.defaults
import esgfpid.exceptions
from esgfpid.utils import loginfo, logdebug, logtrace, logerror, logwarn, log_every_x_times
from naturalsorting import natural_keys

LOGGER = logging.getLogger(__name__)
LOGGER.addHandler(logging.NullHandler())

class NodeManager(object):

    def __init__(self):

        # Props for basic_publish (needed by thread_feeder)
        self.__properties = pika.BasicProperties(
            delivery_mode=esgfpid.defaults.RABBIT_DELIVERY_MODE,
            content_type='application/json',
        )

        # Nodes
        self.__trusted_nodes = {}
        self.__open_nodes = {}
        self.__trusted_nodes_archive = {}
        self.__open_nodes_archive = {}

        # Current node
        self.__current_node = None

        # Important info
        self.__has_trusted = False

    def add_trusted_node(self, **kwargs):
        node_info = self.__add_node(self.__trusted_nodes, self.__trusted_nodes_archive, **kwargs)
        self.__has_trusted = True
        logdebug(LOGGER, 'Trusted rabbit: %s', self.__get_node_log_string(node_info))

    def add_open_node(self, **kwargs):
        added = node_info = self.__add_node(self.__open_nodes, self.__open_nodes_archive, **kwargs)
        logdebug(LOGGER, 'Open rabbit: %s', self.__get_node_log_string(node_info))

    def __add_node(self, store_where, store_archive, **kwargs):
        if self.__has_necessary_info(kwargs):
            node_info = copy.deepcopy(kwargs)
            self.__complete_info_dict(node_info, False)
            self.__store_node_info_by_priority(node_info, store_where)
            self.__store_node_info_by_priority(copy.deepcopy(node_info), store_archive)
            #store_where[node_info['priority']].append(node_info)
            #store_archive[node_info['priority']].append(copy.deepcopy(node_info))
            return node_info
        else:
            raise esgfpid.exceptions.ArgumentError('Cannot add this RabbitMQ node. Missing info. Required: username, password, host and exchange_name. Provided: '+str(kwargs))

    def __store_node_info_by_priority(self, node_info, store_where):
        try:
            store_where[node_info['priority']].append(node_info)
        except KeyError:
            store_where[node_info['priority']] = [node_info]

    def __get_node_log_string(self, node_info):
        return ('%s, %s, %s (exchange "%s")' % (node_info['host'], node_info['username'], node_info['password'], node_info['exchange_name']))

    def __has_necessary_info(self, node_info_dict):
        if ('username' in node_info_dict and
           'password' in node_info_dict and
           'host' in node_info_dict and 
           'exchange_name' in node_info_dict and
            node_info_dict['username'] is not None and
            node_info_dict['password'] is not None and
            node_info_dict['host'] is not None and
            node_info_dict['exchange_name'] is not None):
            return True
        else:
            return False

    def __complete_info_dict(self, node_info_dict, is_open):

        # Make pika credentials
        creds = pika.PlainCredentials(
            node_info_dict['username'],
            node_info_dict['password']
        )
        node_info_dict['credentials'] = creds
        if 'priority' in node_info_dict:
            node_info_dict['priority'] = str(node_info_dict['priority'])
        else:
            node_info_dict['priority'] = 'zzzz_last'

        # Get some defaults:
        socket_timeout = esgfpid.defaults.RABBIT_ASYN_SOCKET_TIMEOUT
        connection_attempts = esgfpid.defaults.RABBIT_ASYN_CONNECTION_ATTEMPTS
        retry_delay = esgfpid.defaults.RABBIT_ASYN_CONNECTION_RETRY_DELAY_SECONDS
        
        # Make pika connection params
        # https://pika.readthedocs.org/en/0.9.6/connecting.html
        params = pika.ConnectionParameters(
            host=node_info_dict['host'], # TODO: PORTS ETC.
            credentials=node_info_dict['credentials'],
            socket_timeout=socket_timeout,
            connection_attempts=connection_attempts,
            retry_delay=retry_delay
        )
        node_info_dict['params'] = params

        # Add some stuff
        node_info_dict['is_open'] = is_open
        '''
        https://pika.readthedocs.org/en/0.9.6/connecting.html
        class pika.connection.ConnectionParameters(
            host=None, port=None, virtual_host=None, credentials=None, channel_max=None,
            frame_max=None, heartbeat_interval=None, ssl=None, ssl_options=None,
            connection_attempts=None, retry_delay=None, socket_timeout=None, locale=None,
            backpressure_detection=None)
        '''
        return node_info_dict

    def get_connection_parameters(self):
        if self.__current_node is None:
            self.set_next_host()
        return self.__current_node['params']

    def has_more_urls(self):
        if self.get_num_left_urls() > 0:
            return True
        return False

    def get_num_left_urls(self):
        return len(self.__trusted_nodes) + len(self.__open_nodes)

    def set_next_host(self):

        if len(self.__trusted_nodes) > 0:
            self.__current_node = self.__get_highest_priority_node(self.__trusted_nodes)
            logdebug(LOGGER, 'Selected a trusted node: %s', self.__current_node['host'])

        elif len(self.__open_nodes) > 0:
            self.__current_node = self.__get_highest_priority_node(self.__open_nodes)
            logdebug(LOGGER, 'Selected an open node: %s', self.__current_node['host'])

        else:
            logwarn(LOGGER, 'No RabbitMQ node left to try! Leaving the last one: %s', self.__current_node['host'])

        self.__exchange_name = self.__current_node['exchange_name']

    def __get_highest_priority_node(self, dict_of_nodes):

        # Get highest priority:
        available_priorities = self.__trusted_nodes.keys()
        available_priorities.sort(key=natural_keys)
        current_priority = available_priorities.pop(0)
        list_of_priority_nodes = self.__trusted_nodes[current_priority]

        # Select one of them
        if len(list_of_priority_nodes)==1:
            nexthost = list_of_priority_nodes.pop()
            if len(list_of_priority_nodes)==0:
                self.__trusted_nodes.pop(current_priority)
            return nexthost
        else:
            nexthost = self.__select_and_remove_random_url_from_list(list_of_priority_nodes)
            return nexthost

    ''' This returns always the same. Does not depend on node. '''
    def get_properties_for_message_publications(self):
        return self.__properties

    ''' This modifies the list! '''
    def __select_and_remove_random_url_from_list(self, list_urls):
        num_urls = len(list_urls)
        random_num = random.randint(0,num_urls-1)
        selected_url = list_urls[random_num]
        list_urls.remove(selected_url)
        return selected_url

    def get_exchange_name(self):
        return self.__exchange_name

    '''
    This flag is appended to the routing key 
    so that we can route messages from the untrusted nodes 
    to other queues.

    Note: The binding has to be done in the RabbitMQ exit 
    node (by the consumer).
    '''
    def get_open_word_for_routing_key(self):

        # Message is published via an open node:
        if self.__current_node['is_open'] == True:
            if self.__has_trusted:
                return 'untrusted-fallback'
            else:
                return 'untrusted-only'

        # Message is published via a trusted node:
        elif self.__current_node['is_open'] == False:
            return 'trusted'

        else:
            logerror(LOGGER, 'Problem: Unsure whether the current node is open or not!')
            return 'untrusted-unsure'

    def reset_nodes(self):
        logdebug(LOGGER, 'Resetting hosts...')
        self.__trusted_nodes = copy.deepcopy(self.__trusted_nodes_archive)
        self.__open_nodes = copy.deepcopy(self.__open_nodes_archive)
        self.set_next_host()

    ''' Only for unit testing! '''
    def _get_num_trusted_and_open_nodes(self):
        n_open = len(self.__open_nodes_archive)
        n_trusted = len(self.__trusted_nodes_archive)
        return (n_trusted,n_open)



import json
import esgfpid.defaults
import random
import logging
from esgfpid.utils import loginfo, logdebug, logtrace, logerror, logwarn

LOGGER = logging.getLogger(__name__)
LOGGER.addHandler(logging.NullHandler())

def get_routing_key_and_string_message_from_message_if_possible(msg):

    # Try to convert message to json:
    json_ok = False
    msg_json = None
    msg_string = None
    
    if msg is None:
        raise ValueError('The message that was passed is None.')
    
    # Get JSON from message, if possible!
    if isinstance(msg, basestring):

        try:
            # Valid string message --> JSON
            msg_string = msg
            msg_json = json.loads(msg)
            json_ok = True
            logdebug(LOGGER, 'Message was transformed to json.')
        except ValueError as e:

            # Invalid string message
            loginfo(LOGGER, 'Message seems to be invalid json: %s', msg)
            msg_string = str(msg)
            json_ok = False
    else:
        try:
            # Message is json already.
            msg_string = json.dumps(msg)
            msg_json = msg
            json_ok = True
            logtrace(LOGGER, 'Message was already json.')

        except TypeError as e:
            if 'not JSON serializable' in e.message:

                # Message was whatever.
                msg_string = str(msg)
                json_ok = False
                msg = ('Message was neither JSON nor string and not understandable: %s' % msg_string)
                loginfo(LOGGER, msg)
                raise ValueError(msg)


    # If we succeeded, try to get routing key:
    routing_key = None
    if json_ok:
        try:
            routing_key = msg_json['ROUTING_KEY']
            logtrace(LOGGER, 'Routing key extracted from message.')
        except (KeyError, TypeError) as e:
            logdebug(LOGGER, 'No routing key in message.')
            routing_key = esgfpid.defaults.RABBIT_DEFAULT_ROUTING_KEY
            pass # There is no routing key in the message
    else:
        routing_key = esgfpid.defaults.RABBIT_DEFAULT_ROUTING_KEY

    return routing_key, msg_string






# List of prefixes
ACCEPTED_PREFIXES = ['21.14100','21.14foo']


# Custom logging
LOG_INFO_TO_DEBUG=False # info messages are printed as "debug" to not appear on publisher console. Should be False, unless you want to almost silence it.
LOG_DEBUG_TO_INFO=False # debug messages are printed as "info" to appear on publisher console. Should be False, except for during tests.
LOG_TRACE_TO_DEBUG=False # Print extremely detailed log messages of rabbit module as "DEBUG" messages. Should be False, except for during tests.
LOG_SHOW_TO_INFO=False # So I can selectively show some log messages to Katharina without having to modify the whole code. Deprecated.

# Solr:
SOLR_HTTPS_VERIFY_DEFAULT=False
SOLR_QUERY_DISTRIB=False

# Rabbit
RABBIT_IS_ASYNCHRONOUS = True
ROUTING_KEY_BASIS = 'cmip6.publisher.HASH.'
RABBIT_DEFAULT_ROUTING_KEY=ROUTING_KEY_BASIS+'fallback' # Default, if none is included in message
RABBIT_EMERGENCY_ROUTING_KEY='UNROUTABLE' # If the message was returned as unroutable by the sender
_persistent = 2
_nonpersistent = 1
RABBIT_DELIVERY_MODE = _persistent # 'delivery_mode': See https://pika.readthedocs.org/en/0.9.6/examples/comparing_publishing_sync_async.html#comparing-message-publishing-with-blockingconnection-and-selectconnection
RABBIT_MANDATORY_DELIVERY = True  # 'mandatory':  "This flag tells the server how to react if the message cannot be routed to a queue. If this flag is set, the server will return an unroutable message with a Return method. If this flag is zero, the server silently drops the message." # See: http://www.rabbitmq.com/amqp-0-9-1-reference.html#basic.publish
RABBIT_FALLBACK_EXCHANGE_NAME = "FALLBACK"

# Synchronous rabbit module:
RABBIT_SYN_SOCKET_TIMEOUT=2 # defaults to 0.25 sec
RABBIT_SYN_CONNECTION_ATTEMPTS=3 # defaults to 1
RABBIT_SYN_CONNECTION_RETRY_DELAY_SECONDS=1 # no default
RABBIT_SYN_MAX_TRIES=3
RABBIT_SYN_TIMEOUT_MILLISEC=10

# Asynchronous rabbit module:
RABBIT_ASYN_RECONNECTION_SECONDS=2 # after how much time try to reconnect if connection was closed?
RABBIT_ASYN_RECONNECTION_MAX_TRIES=60 # how many times should the module try reconnecting? server reboot can take quite a while!
RABBIT_ASYN_SOCKET_TIMEOUT=2 # defaults to 0.25 sec
RABBIT_ASYN_CONNECTION_ATTEMPTS=1 # defaults to 1
RABBIT_ASYN_CONNECTION_RETRY_DELAY_SECONDS=1 # no default
RABBIT_ASYN_FINISH_MAX_TRIES=10 # How many times to recheck if all messages are published+confirmed (on finish)
RABBIT_ASYN_FINISH_WAIT_SECONDS=0.5 # How much time to wait until you recheck (on finish)

# Other
RABBIT_LOG_MESSAGE_INCREMENT = 10



import unittest
import mock
import logging
import json
import requests
import os
import esgfpid.rabbit.asynchronous
import tests.mocks.responsemock
import mock

# Logging:
LOGGER = logging.getLogger(__name__)
LOGGER.addHandler(logging.NullHandler())


class RabbitAsynConnectorTestCase(unittest.TestCase):

    def setUp(self):
        LOGGER.info('######## Next test ########## (%s)' % os.path.basename(__file__))

    def tearDown(self):
        LOGGER.info('#############################')

    def make_testrabbit(self):
        cred = esgfpid.rabbit.connparams.get_credentials('userfoo', 'passwordfoo')
        testrabbit = esgfpid.rabbit.asynchronous.AsynchronousRabbitConnector(
            exchange_name = 'foo_exchange',
            url_preferred = 'first.host.com',
            urls_fallback = ['foo.com', 'bar.de'],
            credentials = cred
        )
        return testrabbit

    # Actual tests:

    def test_init_ok(self):

        # Test variables:
        cred = esgfpid.rabbit.connparams.get_credentials('userfoo', 'passwordfoo')

        # Run code to be tested:
        testrabbit = esgfpid.rabbit.asynchronous.AsynchronousRabbitConnector(
            exchange_name = 'foo_exchange',
            url_preferred = 'first.host.com',
            urls_fallback = ['foo.com', 'bar.de'],
            credentials = cred
        )

        # Check result:
        self.assertIsInstance(testrabbit, esgfpid.rabbit.asynchronous.AsynchronousRabbitConnector, 'Constructor fail.')
        self.assertTrue(testrabbit._AsynchronousRabbitConnector__not_started_yet)


    #
    # Sending messages
    #

    def test_send_message_not_started_yet(self):
        testrabbit = self.make_testrabbit()

        self.assertTrue(testrabbit._AsynchronousRabbitConnector__not_started_yet)

        with self.assertRaises(ValueError):
            testrabbit.send_message_to_queue('message-foo')

        with self.assertRaises(ValueError):
            testrabbit.send_many_messages_to_queue(['a','b','c'])

    @mock.patch('esgfpid.rabbit.asynchronous.asynchronous.AsynchronousRabbitConnector._AsynchronousRabbitConnector__create_thread')
    def test_send_message_ok(self, createpatch):

        # Prepare patch
        threadpatch = mock.MagicMock()
        createpatch.return_value = threadpatch

        # Preparations
        testrabbit = self.make_testrabbit()

        # Run code to be tested:
        testrabbit.start_rabbit_thread()
        testrabbit.send_message_to_queue('foo')
        testrabbit.send_many_messages_to_queue(['a','b','c'])

        # Check result
        threadpatch.start.assert_called_with()
        threadpatch.send_a_message.assert_called_with('foo')
        threadpatch.send_many_messages.assert_called_with(['a','b','c'])

    #
    # Gently finish
    #

    @mock.patch('esgfpid.rabbit.asynchronous.asynchronous.AsynchronousRabbitConnector._AsynchronousRabbitConnector__create_thread')
    def test_gently_finish_ok(self, createpatch):

        # Prepare patch
        threadpatch = mock.MagicMock()
        createpatch.return_value = threadpatch
        threadpatch.is_alive.return_value = False

        # Preparations
        testrabbit = self.make_testrabbit()
        testrabbit.start_rabbit_thread()

        # Run code to be tested:
        testrabbit.finish_rabbit_thread()
        is_done = testrabbit.is_finished()
        any_left = testrabbit.any_leftovers()

        # Check result
        threadpatch.finish_gently.assert_called_with()
        self.assertTrue(is_done)
        self.assertFalse(any_left)

    @mock.patch('esgfpid.rabbit.asynchronous.asynchronous.AsynchronousRabbitConnector._AsynchronousRabbitConnector__create_thread')
    def test_finish_reconnect_send_ok(self, createpatch):

        # Prepare patch
        threadpatch = mock.MagicMock()
        createpatch.return_value = threadpatch
        threadpatch.is_alive.return_value = False

        # Preparations
        testrabbit = self.make_testrabbit()
        testrabbit.start_rabbit_thread()

        # Run code to be tested:
        testrabbit.finish_rabbit_thread()
        testrabbit.reconnect()
        testrabbit.send_message_to_queue('foo')
        testrabbit.send_many_messages_to_queue(['a','b','c'])

        # Check result
        threadpatch.finish_gently.assert_called_with()
        threadpatch.send_a_message.assert_called_with('foo')
        threadpatch.send_many_messages.assert_called_with(['a','b','c'])

    @mock.patch('esgfpid.rabbit.asynchronous.asynchronous.AsynchronousRabbitConnector._AsynchronousRabbitConnector__create_thread')
    def test_rescue_get_ok(self, createpatch):
        '''
        We finish the thread.
        Let's see if the leftovers (that we add articifially) can be retrieved!
        '''

        # Prepare patch
        threadpatch = mock.MagicMock()
        createpatch.return_value = threadpatch
        threadpatch.is_alive.return_value = False
        threadpatch.get_unpublished_messages_as_list.return_value = ['a','b','c']
        threadpatch.get_unconfirmed_messages_as_list.return_value = ['d','e','f']
        threadpatch.get_nacked_messages_as_list.return_value = ['g','h','i']

        # Preparations
        testrabbit = self.make_testrabbit()
        testrabbit.start_rabbit_thread()

        # Run code to be tested:
        testrabbit.finish_rabbit_thread()
        leftovers = testrabbit.get_leftovers()

        # Check result:
        expected_leftovers = ['a','b','c','d','e','f','g','h','i']
        self.assertEquals(leftovers, expected_leftovers, 'Leftovers: %s' % leftovers)


    @mock.patch('esgfpid.rabbit.asynchronous.asynchronous.AsynchronousRabbitConnector._AsynchronousRabbitConnector__create_thread')
    def test_rescue_republish_ok(self, createpatch):
        '''
        We finish and reconnect.
        Let's see if the leftovers (that we add articifially) are published upon reconnect!
        '''

        # Prepare patch
        threadpatch = mock.MagicMock()
        createpatch.return_value = threadpatch
        threadpatch.is_alive.return_value = False
        threadpatch.get_unpublished_messages_as_list.return_value = ['a','b','c']
        threadpatch.get_unconfirmed_messages_as_list.return_value = ['d','e','f']
        threadpatch.get_nacked_messages_as_list.return_value = ['g','h','i']

        # Preparations
        testrabbit = self.make_testrabbit()
        testrabbit.start_rabbit_thread()

        # Run code to be tested:
        testrabbit.finish_rabbit_thread()
        any_left = testrabbit.any_leftovers()
        testrabbit.reconnect()

        # Check result:
        self.assertTrue(any_left)
        threadpatch.send_many_messages.assert_any_call(['g','h','i'])
        threadpatch.send_many_messages.assert_any_call(['d','e','f'])
        threadpatch.send_many_messages.assert_any_call(['a','b','c'])


    if False: # This test takes time!!
        @mock.patch('esgfpid.rabbit.asynchronous.asynchronous.AsynchronousRabbitConnector._AsynchronousRabbitConnector__create_thread')
        def test_join_fails(self, createpatch):
            '''
            We finish, but the join fails.
            This test takes time!
            Rescueing messages should not work, so getting the leftovers should return an empty list.
            '''

            # Prepare patch
            threadpatch = mock.MagicMock()
            createpatch.return_value = threadpatch
            threadpatch.is_alive.return_value = True
            threadpatch.get_unpublished_messages_as_list.return_value = ['a','b','c']
            threadpatch.get_unconfirmed_messages_as_list.return_value = ['d','e','f']
            threadpatch.get_nacked_messages_as_list.return_value = ['g','h','i']

            # Preparations
            testrabbit = self.make_testrabbit()
            testrabbit.start_rabbit_thread()

            # Run code to be tested:
            testrabbit.finish_rabbit_thread()
            leftovers = testrabbit.get_leftovers()

            # Check result:
            self.assertEquals(leftovers, [], 'Leftovers: %s' % leftovers)

    @mock.patch('esgfpid.rabbit.asynchronous.asynchronous.AsynchronousRabbitConnector._AsynchronousRabbitConnector__create_thread')
    def test_force_finish_ok(self, createpatch):
        '''
        We force finish.
        Let's see if the leftovers (that we add articifially) can be retrieved!
        '''

        # Prepare patch
        threadpatch = mock.MagicMock()
        createpatch.return_value = threadpatch
        threadpatch.is_alive.return_value = False
        threadpatch.get_unpublished_messages_as_list.return_value = ['a','b','c']
        threadpatch.get_unconfirmed_messages_as_list.return_value = ['d','e','f']
        threadpatch.get_nacked_messages_as_list.return_value = ['g','h','i']

        # Preparations
        testrabbit = self.make_testrabbit()
        testrabbit.start_rabbit_thread()

        # Run code to be tested:
        testrabbit.force_finish_rabbit_thread()
        leftovers = testrabbit.get_leftovers()

        # Check result:
        expected_leftovers = ['a','b','c','d','e','f','g','h','i']
        self.assertEquals(leftovers, expected_leftovers, 'Leftovers: %s' % leftovers)


import unittest
import mock
import logging
import json
import sys
import esgfpid.assistant.publish
import tests.utils
from tests.utils import compare_json_return_errormessage as error_message
import tests.mocks.rabbitmock
import tests.mocks.solrmock
from esgfpid.defaults import ROUTING_KEY_BASIS as ROUTING_KEY_BASIS
import esgfpid.utils

# Logging
LOGGER = logging.getLogger(__name__)
LOGGER.addHandler(logging.NullHandler())

# Test resources:
from resources.TESTVALUES import TESTVALUES


class ApiTestCase(unittest.TestCase):

    def setUp(self):
        LOGGER.info('######## Next test (%s) ##########', __name__)
        self.make_mocks()

    def tearDown(self):
        LOGGER.info('#############################')

    def make_mocks(self):
        self.default_solrmock = mock.MagicMock()
        self.default_solrmock.is_switched_off.return_value = False
        self.default_rabbitmock = mock.MagicMock()

    def make_patched_connector(self):
        testconnector = esgfpid.Connector(
            handle_prefix = TESTVALUES['prefix'],
            messaging_service_urls_open = TESTVALUES['url_rabbit_open'],
            messaging_service_exchange_name = TESTVALUES['rabbit_exchange_name'],
            data_node = TESTVALUES['data_node'],
            thredds_service_path = TESTVALUES['thredds_service_path'],
            solr_url = TESTVALUES['solr_url'],
            messaging_service_username_open = TESTVALUES['rabbit_username_open']
        )
        # Replace objects that interact with servers with mocks
        self.__patch_connector_with_solr_mock(testconnector)
        self.__patch_connector_with_rabbit_mock(testconnector)
        return testconnector

    def __patch_connector_with_solr_mock(self, connector, solrmock=None):
        if solrmock is None:
            solrmock = self.default_solrmock
        connector._Connector__coupler._Coupler__solr_sender = solrmock

    def __patch_connector_with_rabbit_mock(self, connector, rabbitmock=None):
        if rabbitmock is None:
            rabbitmock = self.default_rabbitmock
        connector._Connector__coupler._Coupler__rabbit_message_sender = rabbitmock

    # Actual tests:

    #
    # Init
    #

    '''
    Test the constructor, with trusted node.
    '''
    def test_init_trusted_ok(self):

        testconnector = esgfpid.Connector(
            handle_prefix = TESTVALUES['prefix'],
            messaging_service_urls_open = TESTVALUES['url_rabbit_open'],
            messaging_service_url_trusted = TESTVALUES['url_rabbit_trusted'],
            messaging_service_exchange_name = TESTVALUES['rabbit_exchange_name'],
            messaging_service_username_open = TESTVALUES['rabbit_username_open'],
            messaging_service_username_trusted = TESTVALUES['rabbit_username_trusted'],
            messaging_service_password = TESTVALUES['rabbit_password']
        )

        # Did init work?
        self.assertIsInstance(testconnector, esgfpid.Connector)

        # Did the module get the right number of
        # trusted and open rabbit nodes?
        node_manager = testconnector._Connector__coupler._Coupler__rabbit_message_sender._RabbitMessageSender__node_manager
        n_trusted, n_open = node_manager._get_num_trusted_and_open_nodes()
        self.assertEquals(n_trusted, 1)
        self.assertEquals(n_open,1)


    '''
    Test the constructor, with only open nodes.
    '''
    def test_init_open_ok(self):

        testconnector = esgfpid.Connector(
            handle_prefix = TESTVALUES['prefix'],
            messaging_service_urls_open = TESTVALUES['url_rabbit_open'],
            messaging_service_exchange_name = TESTVALUES['rabbit_exchange_name'],
            messaging_service_username_open = TESTVALUES['rabbit_username_open']
        )

        # Did init work?
        self.assertIsInstance(testconnector, esgfpid.Connector)

        # Did the module get the right number of
        # trusted and open rabbit nodes?
        node_manager = testconnector._Connector__coupler._Coupler__rabbit_message_sender._RabbitMessageSender__node_manager
        n_trusted, n_open = node_manager._get_num_trusted_and_open_nodes()
        self.assertEquals(n_trusted, 0)
        self.assertEquals(n_open,1)

    def test_passing_consumer_solr_url_ok(self):

        # Init connector
        testconnector = esgfpid.Connector(
            handle_prefix = TESTVALUES['prefix'],
            messaging_service_urls_open = TESTVALUES['url_rabbit_open'],
            messaging_service_exchange_name = TESTVALUES['rabbit_exchange_name'],
            data_node = TESTVALUES['data_node'],
            thredds_service_path = TESTVALUES['thredds_service_path'], # opt
            solr_url = TESTVALUES['solr_url'],
            messaging_service_username_open = TESTVALUES['rabbit_username_open'],
            consumer_solr_url='fake_solr_whatever'
        )

        # Init dataset wizard
        wizard = testconnector.create_publication_assistant(
            drs_id='foo',
            version_number=2016,
            data_node='bar',
            thredds_service_path='baz',
            is_replica=False
        )

        # Did init work?
        self.assertIsInstance(testconnector, esgfpid.Connector)
        self.assertIsInstance(wizard, esgfpid.assistant.publish.DatasetPublicationAssistant)

        # Was the consumer solr url passed?
        self.assertIsNotNone(testconnector._Connector__consumer_solr_url)
        self.assertIsNotNone(wizard._DatasetPublicationAssistant__consumer_solr_url)

    #
    # Publication
    #

    def test_create_publication_assistant_ok(self):

        # Test variables
        drs_id = TESTVALUES['drs_id1']
        version_number = TESTVALUES['version_number1']
        data_node = TESTVALUES['data_node']
        is_replica = False
        thredds_service_path = TESTVALUES['thredds_service_path']

        # Run code to be tested:
        default_testconnector = self.make_patched_connector()
        wizard = default_testconnector.create_publication_assistant(
            drs_id=drs_id,
            version_number=version_number,
            data_node=data_node,
            thredds_service_path=thredds_service_path,
            is_replica=is_replica
        )

        # Check result:
        self.assertIsInstance(wizard, esgfpid.assistant.publish.DatasetPublicationAssistant)


    def test_create_publication_assistant_missing_thredds(self):

        # Preparations: Make connector without thredds
        testconnector = esgfpid.Connector(
            handle_prefix = TESTVALUES['prefix'],
            messaging_service_urls_open = TESTVALUES['url_rabbit_open'],
            messaging_service_exchange_name = TESTVALUES['rabbit_exchange_name'],
            data_node = TESTVALUES['data_node'],
            solr_url = TESTVALUES['solr_url'],
            messaging_service_username_open = TESTVALUES['rabbit_username_open'],
        )

        # Test variables
        drs_id = TESTVALUES['drs_id1']
        version_number = TESTVALUES['version_number1']
        data_node = TESTVALUES['data_node']
        is_replica = False

        # Run code to be tested:
        with self.assertRaises(esgfpid.exceptions.ArgumentError):
            wizard = testconnector.create_publication_assistant(
                drs_id=drs_id,
                version_number=version_number,
                data_node=data_node,
                is_replica=is_replica
            )

    #
    # Unpublication
    #

    def test_unpublish_one_version_ok(self):

        # Test variables
        drs_id = TESTVALUES['drs_id1']
        version_number = TESTVALUES['version_number1']
        data_node = TESTVALUES['data_node']
  
        # Run code to be tested:
        default_testconnector = self.make_patched_connector()
        default_testconnector.unpublish_one_version(
            drs_id=drs_id,
            version_number=version_number,
            data_node=data_node)

        # Check result:
        expected_rabbit_task = {
            "handle": "hdl:"+TESTVALUES['prefix']+'/afd65cd0-9296-35bc-a706-be98665c9c36',
            "operation": "unpublish_one_version",
            "message_timestamp":"anydate",
            "aggregation_level":"dataset",
            "data_node": TESTVALUES['data_node'],
            "ROUTING_KEY": ROUTING_KEY_BASIS+'unpublication.one',
            "drs_id":drs_id,
            "version_number": int(version_number)
        }
        received_rabbit_task = self.default_rabbitmock.send_message_to_queue.call_args[0][0] # first get positional args, second get the first og those
        tests.utils.replace_date_with_string(received_rabbit_task)
        is_same = tests.utils.is_json_same(expected_rabbit_task, received_rabbit_task)
        self.assertTrue(is_same, tests.utils.compare_json_return_errormessage(expected_rabbit_task, received_rabbit_task))

    def test_unpublish_one_version_with_consumer_url_ok(self):

        # Test variables
        drs_id = TESTVALUES['drs_id1']
        version_number = TESTVALUES['version_number1']
        data_node = TESTVALUES['data_node']

        # Make test connector
        testconnector = esgfpid.Connector(
            handle_prefix = TESTVALUES['prefix'],
            messaging_service_urls_open = TESTVALUES['url_rabbit_open'],
            messaging_service_exchange_name = TESTVALUES['rabbit_exchange_name'],
            data_node = TESTVALUES['data_node'],
            solr_url = TESTVALUES['solr_url'],
            messaging_service_username_open = TESTVALUES['rabbit_username_open'],
            consumer_solr_url="fake_solr_whatever"
        )
        self.__patch_connector_with_rabbit_mock(testconnector)
  
        # Run code to be tested:
        testconnector.unpublish_one_version(
            drs_id=drs_id,
            version_number=version_number,
            data_node=data_node)

        # Check result:
        expected_rabbit_task = {
            "handle": "hdl:"+TESTVALUES['prefix']+'/afd65cd0-9296-35bc-a706-be98665c9c36',
            "operation": "unpublish_one_version",
            "message_timestamp":"anydate",
            "aggregation_level":"dataset",
            "data_node": TESTVALUES['data_node'],
            "ROUTING_KEY": ROUTING_KEY_BASIS+'unpublication.one',
            "drs_id":drs_id,
            "version_number": int(version_number)
        } # We don't get the consumer_solr_url, because it is only needed for all versions.
        received_rabbit_task = self.default_rabbitmock.send_message_to_queue.call_args[0][0] # first get positional args, second get the first og those
        tests.utils.replace_date_with_string(received_rabbit_task)
        is_same = tests.utils.is_json_same(expected_rabbit_task, received_rabbit_task)
        self.assertTrue(is_same, tests.utils.compare_json_return_errormessage(expected_rabbit_task, received_rabbit_task))

    @mock.patch('esgfpid.coupling.Coupler.retrieve_datasethandles_or_versionnumbers_of_allversions')
    def test_unpublish_all_versions_nosolr_ok(self, solr_asker_patch):

        # Patch coupler
        mydict = dict(dataset_handles=None, version_numbers=None)
        solr_asker_patch.return_value = mydict

        # Test variables
        drs_id = TESTVALUES['drs_id1']
        data_node = TESTVALUES['data_node']

        # Run code to be tested:
        default_testconnector = self.make_patched_connector()
        default_testconnector.unpublish_all_versions(
            drs_id=drs_id,
            data_node=data_node)

        # Check result:
        expected_rabbit_task = {
            "operation": "unpublish_all_versions",
            "message_timestamp": "anydate",
            "data_node": TESTVALUES['data_node'],
            "aggregation_level":"dataset",
            "drs_id":drs_id,
            "ROUTING_KEY": ROUTING_KEY_BASIS+'unpublication.all'
        }
        
        received_rabbit_task = self.default_rabbitmock.send_message_to_queue.call_args[0][0] # first get positional args, second get the first og those
        tests.utils.replace_date_with_string(received_rabbit_task)
        is_same = tests.utils.is_json_same(expected_rabbit_task, received_rabbit_task)
        self.assertTrue(is_same, tests.utils.compare_json_return_errormessage(expected_rabbit_task, received_rabbit_task))

    @mock.patch('esgfpid.coupling.Coupler.retrieve_datasethandles_or_versionnumbers_of_allversions')
    def test_unpublish_all_versions_nosolr__butconsumersolr_ok(self, solr_asker_patch):

        # Patch coupler
        mydict = dict(dataset_handles=None, version_numbers=None)
        solr_asker_patch.return_value = mydict

        # Make test connector
        drs_id = TESTVALUES['drs_id1']
        version_number = TESTVALUES['version_number1']
        data_node = TESTVALUES['data_node']
        testconnector = esgfpid.Connector(
            handle_prefix = TESTVALUES['prefix'],
            messaging_service_urls_open = TESTVALUES['url_rabbit_open'],
            messaging_service_exchange_name = TESTVALUES['rabbit_exchange_name'],
            data_node = TESTVALUES['data_node'],
            solr_url = TESTVALUES['solr_url'],
            messaging_service_username_open = TESTVALUES['rabbit_username_open'],
            consumer_solr_url="fake_solr_whatever"
        )
        self.__patch_connector_with_rabbit_mock(testconnector)

        # Test variables
        drs_id = TESTVALUES['drs_id1']
        data_node = TESTVALUES['data_node']

        # Run code to be tested:
        testconnector.unpublish_all_versions(
            drs_id=drs_id,
            data_node=data_node)

        # Check result:
        expected_rabbit_task = {
            "operation": "unpublish_all_versions",
            "message_timestamp": "anydate",
            "data_node": TESTVALUES['data_node'],
            "aggregation_level":"dataset",
            "drs_id":drs_id,
            "ROUTING_KEY": ROUTING_KEY_BASIS+'unpublication.all',
            "consumer_solr_url":"fake_solr_whatever"
        }
        
        received_rabbit_task = self.default_rabbitmock.send_message_to_queue.call_args[0][0] # first get positional args, second get the first og those
        tests.utils.replace_date_with_string(received_rabbit_task)
        is_same = tests.utils.is_json_same(expected_rabbit_task, received_rabbit_task)
        self.assertTrue(is_same, tests.utils.compare_json_return_errormessage(expected_rabbit_task, received_rabbit_task))

    #
    # Errata
    #

    def test_add_errata_id_several_ok(self):

        # Test variables:
        drs_id = TESTVALUES['drs_id1']
        version_number = TESTVALUES['version_number1']
        errata_ids = TESTVALUES['errata_ids']

        # Run code to be tested:
        default_testconnector = self.make_patched_connector()
        default_testconnector.add_errata_ids(
            drs_id=drs_id,
            version_number=version_number,
            errata_ids=errata_ids
        )

        # Check result:
        expected_rabbit_task = {
            "handle": "hdl:"+TESTVALUES['prefix']+"/afd65cd0-9296-35bc-a706-be98665c9c36",
            "operation": "add_errata_ids",
            "message_timestamp":"anydate",
            "errata_ids":errata_ids,
            "ROUTING_KEY": ROUTING_KEY_BASIS+'errata.add',
            "drs_id":drs_id,
            "version_number":version_number
        }

        received_rabbit_task = self.default_rabbitmock.send_message_to_queue.call_args[0][0] # first get positional args, second get the first og those
        tests.utils.replace_date_with_string(received_rabbit_task)
        is_same = tests.utils.is_json_same(expected_rabbit_task, received_rabbit_task)
        self.assertTrue(is_same, tests.utils.compare_json_return_errormessage(expected_rabbit_task, received_rabbit_task))

    def test_add_errata_id_one_ok(self):

        # Test variables:
        drs_id = TESTVALUES['drs_id1']
        version_number = TESTVALUES['version_number1']
        errata_id = TESTVALUES['errata_id']

        # Run code to be tested:
        default_testconnector = self.make_patched_connector()
        default_testconnector.add_errata_ids(
            drs_id=drs_id,
            version_number=version_number,
            errata_ids=errata_id
        )

        # Check result:
        expected_rabbit_task = {
            "handle": "hdl:"+TESTVALUES['prefix']+"/afd65cd0-9296-35bc-a706-be98665c9c36",
            "operation": "add_errata_ids",
            "message_timestamp":"anydate",
            "errata_ids":[errata_id],
            "ROUTING_KEY": ROUTING_KEY_BASIS+'errata.add',
            "drs_id":drs_id,
            "version_number":version_number
        }

        received_rabbit_task = self.default_rabbitmock.send_message_to_queue.call_args[0][0] # first get positional args, second get the first og those
        tests.utils.replace_date_with_string(received_rabbit_task)
        is_same = tests.utils.is_json_same(expected_rabbit_task, received_rabbit_task)
        self.assertTrue(is_same, tests.utils.compare_json_return_errormessage(expected_rabbit_task, received_rabbit_task))

    #
    # Shopping Cart
    #

    def test_make_shopping_cart_pid(self):

        # Test variables
        prefix = TESTVALUES['prefix']
        content1 = {'foo':'foo', 'bar':'bar'}
        content2 = {'foo':'foo', 'bar': None}

        # Run code to be tested:
        default_testconnector = self.make_patched_connector()
        pid1 = default_testconnector.create_shopping_cart_pid(content1)
        received_rabbit_task1 = self.default_rabbitmock.send_message_to_queue.call_args[0][0]
        pid2 = default_testconnector.create_shopping_cart_pid(content2)
        received_rabbit_task2 = self.default_rabbitmock.send_message_to_queue.call_args[0][0]

        # Check result:
        expected_handle_both_cases = "hdl:"+prefix+"/339427df-edbd-3f43-acf2-80ddc7729f27"
        expected_rabbit_task1 = {
            "handle": expected_handle_both_cases,
            "operation": "shopping_cart",
            "message_timestamp":"anydate",
            "data_cart_content":content1,
            "ROUTING_KEY": ROUTING_KEY_BASIS+'cart.datasets'
        }
        expected_rabbit_task2 = {
            "handle": expected_handle_both_cases,
            "operation": "shopping_cart",
            "message_timestamp":"anydate",
            "data_cart_content":content2,
            "ROUTING_KEY": ROUTING_KEY_BASIS+'cart.datasets'
        }
        
        tests.utils.replace_date_with_string(received_rabbit_task1)
        tests.utils.replace_date_with_string(received_rabbit_task2)
        same1 = tests.utils.is_json_same(expected_rabbit_task1, received_rabbit_task1)
        same2 = tests.utils.is_json_same(expected_rabbit_task2, received_rabbit_task2)
        self.assertTrue(same1, error_message(expected_rabbit_task1, received_rabbit_task1))
        self.assertTrue(same2, error_message(expected_rabbit_task2, received_rabbit_task2))
        self.assertTrue(pid1==pid2, 'Both pids are not the same.')

    #
    # Threads
    #

    def test_start_messaging_thread(self):
        LOGGER.debug('Thread test')
        default_testconnector = self.make_patched_connector()
        default_testconnector.start_messaging_thread()
        self.default_rabbitmock.start.assert_called_with()

    def test_finish_messaging_thread(self):
        LOGGER.debug('Thread test')
        default_testconnector = self.make_patched_connector()
        default_testconnector.finish_messaging_thread()
        self.default_rabbitmock.finish.assert_called_with()

    def test_force_finish_messaging_thread(self):
        LOGGER.debug('Thread test')
        default_testconnector = self.make_patched_connector()
        default_testconnector.force_finish_messaging_thread()
        self.default_rabbitmock.force_finish.assert_called_with()
